import pytest
import numpy as np
from pyempfin.xsap import _winsor_njit, _newey_njit, estbeta, estbeta1m
from pyempfin.xsap import format_table, fmreg
from pyempfin.xsap import groupby_wavg, get_port_ret, tscssum
import pandas as pd
from pandas.testing import assert_frame_equal, assert_series_equal
import re

def assert_float_equal(x, y, tol=1e-15):
    return np.abs(x-y) < tol

def checkarr(arr: np.ndarray, orgarr: np.ndarray,
             lower: float, upper: float) -> bool:
    # Same missing values
    cond1 = (np.isfinite(arr) == np.isfinite(orgarr)).all()
    # Check maximum and minimum values
    cond2 = np.nanmax(arr) == upper
    cond3 = np.nanmin(arr) == lower
    # Check other numbers are not modified
    cond4 = (arr[(lower < arr) & (arr < upper)] ==
             orgarr[(lower < orgarr) & (orgarr < upper)]).all()
    return cond1 and cond2 and cond3 and cond4

@pytest.fixture
def xsstk():
    stk = pd.read_csv('xsstktestdata.csv')
    stk.columns = stk.columns.str.lower()
    stk['datemn'] = pd.to_datetime(stk['date'].astype(str)).dt.to_period('M')
    stk = stk.set_index(['permno', 'datemn']).sort_index()
    stk = stk[['exret']]
    return stk

@pytest.fixture
def ff():
    ff = pd.read_csv('ff.csv')
    ff['datemn'] = pd.to_datetime(ff['date'].astype(str)).dt.to_period('M')
    ff.drop(columns=['date'], inplace=True)
    ff.set_index('datemn', inplace=True)
    return ff

def test_winsor():
    # Test array
    arr = np.arange(0, 101)
    # Add missing
    arr = np.hstack((arr, np.array([np.nan]*10)))
    # Shuffle
    np.random.shuffle(arr)
    # Winsorize
    w1 = _winsor_njit(arr, (0.052, 0.959), interpolation='inner')
    assert checkarr(w1, arr, lower=6.0, upper=95.0)
    w2 = _winsor_njit(arr, (0.052, 0.959), interpolation='outer')
    assert checkarr(w2, arr, lower=5.0, upper=96.0)


def test_newey(ff):
    # Load data
    ffres = pd.read_stata('ff_mtkrf_newey.dta')
    data = ff['mktrf'].to_numpy()
    data = data - data.mean()
    # Test
    stderr2 = np.zeros(ffres.shape[0])
    for i in range(ffres.shape[0]):
        stderr2[i] = _newey_njit(data, int(ffres['lag'].iloc[i]))
    assert (np.abs(stderr2 - ffres['stderr'].to_numpy())<1e-5).all()

def test_estbeta1m_fmreg(ff, xsstk):
    # estbeta1m and Stata's rangestat are different. estbeta1m first convert
    # long-form panel data into wide form (rows are periods and columns are
    # stocks) filling missing returns as NaN. Then, estimate beta for each
    # column in rolling windows. As a result, for some stock and some period,
    # even if its return is missing, beta can still be calculated as long as
    # there are enough observations in the look-back period. However in such
    # case, rangestat will not estimate beta. As a result, when comparing the
    # results, we have to remove those cases where rangstat does not calculate.
    # The easiest way is to start from the panel data containing returns and
    # left join with beta estimates.
    # ---------- Test for beta estimation
    model = ['mktrf', 'smb', 'hml', 'umd']
    cols_test = model
    cols_targ = ['b_' + x for x in model]
    beta1 = estbeta1m(
        leftdata=xsstk['exret'],
        rightdata=ff,
        model=model,
        window=(-24, -1),
        minobs=6,
        hasconst=True
    )
    assert not isinstance(beta1.columns, pd.MultiIndex)
    beta2 = estbeta1m(
        leftdata=xsstk['exret'],
        rightdata=ff,
        model=model,
        window=(-20, -5),
        minobs=10,
        hasconst=True
    )
    assert not isinstance(beta2.columns, pd.MultiIndex)
    beta3 = estbeta1m(
        leftdata=xsstk['exret'],
        rightdata=ff,
        model=model,
        window=(-12, 0),
        minobs=13,
        hasconst=True
    )
    assert not isinstance(beta3.columns, pd.MultiIndex)
    betares1 = pd.read_stata('testbetares1.dta')
    betares1['datemn'] = betares1['datemn'].dt.to_period('M')
    betares1.set_index(['permno', 'datemn'], inplace=True)
    betares2 = pd.read_stata('testbetares2.dta')
    betares2['datemn'] = betares2['datemn'].dt.to_period('M')
    betares2.set_index(['permno', 'datemn'], inplace=True)
    betares3 = pd.read_stata('testbetares3.dta')
    betares3['datemn'] = betares3['datemn'].dt.to_period('M')
    betares3.set_index(['permno', 'datemn'], inplace=True)
    comp1 = xsstk[['exret']].join(beta1[cols_test]).join(betares1[cols_targ])
    comp2 = xsstk[['exret']].join(beta2[cols_test]).join(betares2[cols_targ])
    comp3 = xsstk[['exret']].join(beta3[cols_test]).join(betares3[cols_targ])
    # Check missing values
    assert comp1[cols_test+cols_targ].isna().sum().std() == 0
    assert comp2[cols_test+cols_targ].isna().sum().std() == 0
    assert comp3[cols_test+cols_targ].isna().sum().std() == 0
    # Drop NA
    comp1.dropna(subset=cols_test + cols_targ, inplace=True)
    comp2.dropna(subset=cols_test + cols_targ, inplace=True)
    comp3.dropna(subset=cols_test + cols_targ, inplace=True)
    # Check values
    assert (np.abs(
        comp1[cols_test].to_numpy() - comp1[cols_targ].to_numpy()
    ) < 1e-5).all()
    assert (np.abs(
        comp2[cols_test].to_numpy() - comp2[cols_targ].to_numpy()
    ) < 1e-5).all()
    assert (np.abs(
        comp3[cols_test].to_numpy() - comp3[cols_targ].to_numpy()
    ) < 1e-5).all()
    # ---------- Test for Fama-MacBeth regression
    models = [
        ['exret', 'mktrf'],
        ['exret', 'mktrf', 'smb', 'hml'],
        ['exret', 'mktrf', 'smb', 'hml', 'umd'],
    ]
    # Test result
    fmres1 = fmreg(
        leftdata=comp1[['exret']],
        rightdata=comp1[cols_test],
        models=models,
        maxlag=5,
        roworder=['mktrf', 'smb', 'hml', 'umd'],
        hasconst=True,
        scale=1,
        getlambda=False,
        estfmt=('.7f', '.7f'),
    )
    # Result from Stata's xtfmb
    fmres1c = pd.read_stata('testfmreg1.dta')
    fmres1c['model'] = fmres1c['model'].astype('int') - 1
    fmres1c['indepvar'] = fmres1c['indepvar'].replace({'_cons': 'Constant'})
    fmres1c['indepvar'] = fmres1c['indepvar'].str.replace('b_', '')
    fmres1c.set_index(['model', 'indepvar'], inplace=True)
    # Test
    for m in range(len(models)):
        indepvars = models[m][1:]
        for v in indepvars + ['Constant']:
            i = fmres1[''].to_list().index(v)
            print(f'{i}, {v}')
            # Test coefficient
            assert np.abs(
                fmres1c.loc[m].loc[v]['b'] -
                float(re.match('^([0-9\.\-]*)', fmres1[m].iloc[i]).group(1))
            ) < 1e-6
            # Test t-stat
            assert np.abs(
                fmres1c.loc[m].loc[v]['t'] -
                float(re.match(r'^\(([0-9\.\-]*)\)', fmres1[m].iloc[i+1]).group(1))
            ) < 1e-6
    # Test scale
    fmres1 = fmreg(
        leftdata=comp1[['exret']],
        rightdata=comp1[cols_test],
        models=models,
        maxlag=5,
        roworder=['mktrf', 'smb', 'hml', 'umd'],
        hasconst=True,
        scale=100,
        getlambda=False,
        estfmt=('.7f', '.7f'),
    )
    for m in range(len(models)):
        indepvars = models[m][1:]
        for v in indepvars + ['Constant']:
            i = fmres1[''].to_list().index(v)
            print(f'{i}, {v}')
            # Test coefficient
            assert np.abs(
                fmres1c.loc[m].loc[v]['b']*100 -
                float(re.match('^([0-9\.\-]*)', fmres1[m].iloc[i]).group(1))
            ) < 1e-6
            # Test t-stat
            assert np.abs(
                fmres1c.loc[m].loc[v]['t'] -
                float(re.match(r'^\(([0-9\.\-]*)\)', fmres1[m].iloc[i+1]).group(1))
            ) < 1e-6


def test_estbeta(ff, xsstk):
    # Index
    models = {
        'm1': ['exret', 'mktrf'],
        'm2': ['exret', 'mktrf', 'smb', 'hml'],
        'm3': ['exret', 'mktrf', 'smb', 'hml', 'umd']
    }
    beta1 = estbeta(
        leftdata=xsstk,
        rightdata=ff,
        models=models,
        window=(-20, -5),
        minobs=10,
        hasconst=True
    )
    assert beta1.columns.nlevels == 2
    beta1.columns = beta1.columns.to_flat_index()
    betares1 = pd.read_stata('testbetares4.dta')
    betares1['datemn'] = betares1['datemn'].dt.to_period('M')
    betares1.set_index(['permno', 'datemn'], inplace=True)
    l1 = beta1.columns.tolist()
    l2 = betares1.columns.tolist()
    c1 = xsstk[['exret']].join(beta1) \
        .join(betares1) \
        .dropna(subset=l1 + l2)
    assert (np.abs(c1[l1].to_numpy() - c1[l2].to_numpy()) < 1e-5).all()

    # Let stk and ff have a MultiIndex
    xsstk2 = xsstk.copy()
    xsstk2.columns = pd.MultiIndex.from_arrays([['exret'], ['exret']])
    ff2 = ff.copy()
    ff2.columns = pd.MultiIndex.from_product([['m1'], ff2.columns])
    ff2 = ff2.join(
        ff2.set_axis(ff2.columns.set_levels(['m2'], level=0), axis='columns'))
    models2 = {
        'm1': [('exret', 'exret'), ('m1', 'mktrf')],
        'm2': [('exret', 'exret'), ('m1', 'mktrf'), ('m1', 'smb'), ('m1', 'hml')],
        'm3': [('exret', 'exret'), ('m1', 'mktrf'), ('m1', 'smb'), ('m1', 'hml'),
               ('m1', 'umd')]
    }
    beta2 = estbeta(
        leftdata=xsstk2,
        rightdata=ff2,
        models=models2,
        window=(-20, -5),
        minobs=10,
        hasconst=True
    )
    l3 = ['_' + x for x in l2]
    c2 = xsstk[['exret']].join(beta2.set_axis(l3, axis='columns')) \
        .join(betares1) \
        .dropna(subset=l3 + l2)
    assert (np.abs(c2[l3].to_numpy() - c2[l2].to_numpy()) < 1e-5).all()



def test_format_table():
    df = pd.DataFrame({
        'intcol': [1000, 101, 2123432, 0],
        'floatcol': [0.1234934, 0.2342341234, 0.24238, 0.23427423],
    }, index=[f'Row {i}' for i in range(4)])
    sumstat = format_table(df, 4)
    df_corr = pd.DataFrame({
        'intcol': ['1,000', '101', '2,123,432', '0'],
        'floatcol': ['0.1235', '0.2342', '0.2424', '0.2343'],
    }, index=df.index)
    assert_frame_equal(sumstat, df_corr)


def test_groupby_wavg():
    N = 1000
    np.random.seed(1234)
    df = pd.DataFrame({
        'byvar1': np.random.randint(0, 5, (N,)),
        'byvar2': np.random.randint(0, 10, (N,)),
        'x': np.random.randn(N),
        'w': np.random.rand(N),
    }).sort_values(['byvar1', 'byvar2']).reset_index(drop=True)
    df.loc[np.random.randint(0, N, (int(.1*N),)), 'x'] = np.nan
    df.loc[np.random.randint(0, N, (int(.1*N),)), 'w'] = np.nan
    dftest = groupby_wavg(
        data=df, bys=['byvar1', 'byvar2'], var='x', weight='w'
    )
    dfcorr = df.dropna(subset=['x', 'w']).groupby(['byvar1', 'byvar2']).apply(
        lambda x: np.average(x['x'], weights=x['w'])
    )
    assert_series_equal(dftest, dfcorr, check_names=False)


def test_tscssum():
    N = 1000
    np.random.seed(1234)
    df = pd.DataFrame({
        'byvar1': np.random.randint(0, 5, (N,)),
        'byvar2': np.random.randint(0, 10, (N,)),
        'x': np.random.randn(N),
        'y': np.random.rand(N),
        'z': np.random.rand(N),
    }).sort_values(['byvar1', 'byvar2']).reset_index(drop=True)
    df.loc[np.random.randint(0, N, (int(.1*N),)), 'x'] = np.nan
    df.loc[np.random.randint(0, N, (int(.1*N),)), 'y'] = np.nan
    df.loc[np.random.randint(0, N, (int(.1*N),)), 'z'] = np.nan
    dftest = tscssum(
        df, by=['byvar1', 'byvar2'], subset=['x', 'z'], percentiles=(.01, .05, .25, .50, .75, .95, .99)
    )
    for v in ['x', 'z']:
        assert_float_equal(
            dftest.loc[v, 'N'],
            df.groupby(['byvar1','byvar2'])[v].count().sum()
        )
        assert_float_equal(
            dftest.loc[v, 'Mean'],
            df.groupby(['byvar1','byvar2'])[v].mean().mean()
        )
        assert_float_equal(
            dftest.loc[v, 'Std'],
            df.groupby(['byvar1','byvar2'])[v].std().mean()
        )
        assert_float_equal(
            dftest.loc[v, 'Min'],
            df.groupby(['byvar1','byvar2'])[v].min().mean()
        )
        assert_float_equal(
            dftest.loc[v, 'Max'],
            df.groupby(['byvar1','byvar2'])[v].max().mean()
        )
        for pct in [1, 5, 25, 50, 75, 95, 99]:
            assert_float_equal(
                dftest.loc[v, f'p{pct}'],
                df.groupby(['byvar1', 'byvar2'])[v].quantile(pct/100).mean(),
            )


# def test_get_port_ret():
#     N = 1000
#     np.random.seed(1234)
#     df = pd.DataFrame({
#         'period': np.random.randint(0, 5, (N,)),
#         'sortvar1': np.random.rand(N),
#         'sortvar2': np.random.rand(N),
#         'x': np.random.randn(N),
#         'w': np.random.rand(N),
#     }).sort_values(['sortvar1', 'sortvar2']).reset_index(drop=True)
#     df.loc[np.random.randint(0, N, (int(.1*N),)), 'sortvar1'] = np.nan
#     df.loc[np.random.randint(0, N, (int(.1*N),)), 'sortvar2'] = np.nan
#     df.loc[np.random.randint(0, N, (int(.1*N),)), 'x'] = np.nan
#     df.loc[np.random.randint(0, N, (int(.1*N),)), 'w'] = np.nan
#     dftest = get_port_ret(
#         data=df, nq=5, timevar='period', retvar='x',
#         rnkvars=['sortvar1', 'sortvar2'], rnkvarnames=['rnk1', 'rnk2'],
#         wvar='w', dep=True,
#     )
#



# if __name__ == '__main__':
#     arr = np.arange(0, 101)
#     _winsor_njit(arr, (0.05, 0.95), interpolation='inner')


import numpy as np
import pandas as pd
from functools import partial
from typing import Union, List, Tuple
from numba import njit, generated_jit
import numba


ARR_TYPES = (numba.types.Array)


def winsor(datacol: pd.Series, lower: float, upper: float) -> pd.Series:
    """Winsorize the pandas series at the given percentiles

    Parameters
    ----------
    datacol : pandas.Series
    lower : float
        Lower percentile to winsorize. The value should be between 0 and 1
    upper : float
        Upper percentile to winsorize. The value should be between 0 and 1

    Returns
    -------
    pandas.Series
    """
    assert 0 <= lower <= 1, 'Lower percentile is not between 0 and 1'
    assert 0 <= upper <= 1, 'Upper percentile is not between 0 and 1'
    return datacol.clip(
        lower=np.nanquantile(datacol, q=lower),
        upper=np.nanquantile(datacol, q=upper)
    )


def trunc(data: np.ndarray, lower: float, upper: float) -> np.ndarray:
    """Truncate data at specific percentiles
        E.g. _trunc(data, threshold=(0.01, 0.99)) truncate data
        at 1% and 99%.    

    Parameters
    ----------
    data : np.ndarray
    lower : float
        Lower percentile to truncate. The value should be between 0 and 1
    upper : float
        Upper percentile to truncate. The value should be between 0 and 1

    Returns
    -------
    np.ndarray
    """    
    assert 0 <= lower <= 1, 'Lower percentile is not between 0 and 1'
    assert 0 <= upper <= 1, 'Upper percentile is not between 0 and 1'
    # Get mask for non-missing values
    nnamask = ~np.isnan(data)
    # Get lower and upper threshold
    lo = np.quantile(data[nnamask], lower)
    up = np.quantile(data[nnamask], upper)
    # Get values between lo and up
    res = np.nan * np.ones(len(data))
    mask = (lo <= data) & (data <= up)
    res[mask] = data[mask]
    return res


def sumstat(data: Union[pd.DataFrame,pd.Series], subset: Union[list,None]=None,
            percentiles: tuple=(.01, .05, .50, .95, .99)) -> pd.DataFrame:
    """Print summary statistics

    Parameters
    ----------
    data : pandas.DataFrame or pandas.Series
        The data table to generate summary statistics
    subset : list of str, default: None
        The list of column names to generate summary statistics. If None, all
        columns are used.
    percentiles : tuple of float, default: (.01, .05, .25, .50, .75, .95, .99)
        The list of percentiles in the table.

    Returns
    -------
    pandas.DataFrame or pandas.Series
        The summary statistics. Each row represents one variable and each
        column represents one statistic.
    """
    funclist = [
                   lambda x: np.isfinite(x).sum(),
                   np.nanmean,
                   np.nanstd,
                   np.nanmin
               ] + [partial(lambda x, i: np.nanpercentile(x, q=i * 100), i=i)
                    for i in percentiles
                    ] + [np.nanmax]
    if isinstance(data, pd.DataFrame):
        if subset is None:
            subset = data.columns
        res = pd.DataFrame(
            [[func(data[x].to_numpy()) for func in funclist] for x in subset],
            index=subset,
            columns=[
                        'N', 'Mean', 'Std', 'Min'
                    ] + ['p' + str(int(i * 100)) for i in percentiles] + ['Max']
        )
        return res
    elif isinstance(data, pd.Series):
        res = pd.Series(
            [func(data.to_numpy()) for func in funclist],
            index=[
                        'N', 'Mean', 'Std', 'Min'
                    ] + ['p' + str(int(i * 100)) for i in percentiles] + ['Max']
        )
        return res
    else:
        raise ValueError('data is not of pd.DataFrame or pd.Series type!')



def ndup(data: pd.DataFrame, subset: Union[list,None]=None) -> int:
    """Number of duplicates over the subset of variables

    Parameters
    ----------
    data : pandas.DataFrame
    subset : list of str, default: None

    Returns
    -------
    int
    """
    # If subset=None, duplicated uses all columns by default
    return data.loc[lambda x: x.duplicated(subset=subset)].shape[0]


def getdup(data: pd.DataFrame, subset: Union[list,None]=None) -> pd.DataFrame:
    """Return rows of duplicated values over the subset of variables

    Parameters
    ----------
    data : pandas.DataFrame
    subset : list of str, default: None

    Returns
    -------
    pandas.DataFrame
    """
    return data.loc[lambda x: x.duplicated(subset=subset, keep=False)]



def getnodup(data: pd.DataFrame, subset: Union[list,None]=None) -> pd.DataFrame:
    """Return rows without duplicated values over the subset of variables

    Parameters
    ----------
    data : pandas.DataFrame
    subset : list of str, default: None

    Returns
    -------
    pandas.DataFrame
    """
    return data.loc[lambda x: ~x.duplicated(subset=subset, keep=False)]



# Count missing values
def desmiss(data: pd.DataFrame,
            subset: Union[list,None]=None,
            not_missing: bool=False,
            remove_zero: bool=False,
            sort: Union[str,None]='descending') -> pd.DataFrame:
    """Return a description of missing (or non-missing) observations in the data set

    Parameters
    ----------
    data : pandas.DataFrame
    subset : list of str, default: None
    not_missing : bool, default: False
    remove_zero : bool, default: False
    sort : {'descending', 'ascending', None}, default: 'descending'

    Returns
    -------
    pandas.DataFrame
    """
    if subset is None:
        subset = data.columns
    if not_missing:
        out = pd.DataFrame({
            '# not NaN': data[subset].apply(lambda x: (~_is_missing(x)).sum())
        })
        out['% not NaN'] = out['# not NaN'].apply(
            lambda x: '{:.2f}%'.format(x / data.shape[0] * 100)
        )
        sort_var = '# not NaN'
    else:
        out = pd.DataFrame({
            '# NaN': data[subset].apply(lambda x: _is_missing(x).sum())
        })
        out['% NaN'] = out['# NaN'].apply(
            lambda x: '{:.2f}%'.format(x / data.shape[0] * 100)
        )
        out = out.sort_values('# NaN', ascending=False)
        sort_var = '# NaN'
    # Sort
    if sort == 'descending':
        out = out.sort_values(sort_var, ascending=False)
    elif sort == 'ascending':
        out = out.sort_values(sort_var, ascending=True)
    elif sort is None:
        pass
    else:
        raise ValueError('Invalid value for argument sort')
    # Keep all or not
    if remove_zero:
        out = out.loc[lambda x: x.max(axis=1) > 0]
    return out

def _is_missing(data: Union[pd.Series, pd.DataFrame]) -> Union[pd.Series, pd.DataFrame]:
    return data.isna() | (data == '')

# Mimic the coalesce function in SAS
def coalesce(*argv) -> pd.Series:
    """Mimic the coalesce function in SAS.

    Parameters
    ----------
    *argv : pandas.Series or a number
        Each argument is a pandas Series or a number.

    Returns
    -------
    pandas.Series
    """
    narg = len(argv)
    out = argv[0]
    for i in range(narg - 1):
        # Construct a constant series if a number is passed
        if isinstance(argv[i + 1], pd.Series):
            out = out.combine_first(argv[i + 1])
        else:
            out = out.combine_first(
                pd.Series([argv[i + 1]] * out.shape[0], index=out.index)
            )
    return out


def weighted_mean(data: pd.DataFrame,
                  by: list,
                  value: str,
                  weight: str,
                  wavg: str) -> pd.DataFrame:
    """Calculate weighted average

    Parameters
    ----------
    data : pd.DataFrame
        Input DataFrame
    by : list of str
        List of column names to group by
    value : str
        Column name of the variable to calculate average
    weight : str
        Column name of the variable as weight
    wavg : str
        Column name of the calculated weighted average

    Returns
    -------
    pd.DataFrame
        Output DataFrame with column names in by and wavg.
    """
    raise DeprecationWarning('Note that weighted_mean is slower than groupby_wavg')
    # Dict to encode column names
    encodedict = {by[i]: f'by_{i}' for i in range(len(by))}
    encodedict[value] = 'value'
    encodedict[weight] = 'weight'
    # Dict to decode column names
    decodedict = {f'by_{i}': by[i] for i in range(len(by))}
    decodedict['value'] = value
    decodedict['weight'] = weight
    decodedict['wavg'] = wavg
    # Assign encoded names
    by_ = [f'by_{i}' for i in range(len(by))]
    # Select useful columns
    data1 = data[by + [value, weight]]
    # Encode column names
    data1.columns = data1.columns.to_series().replace(encodedict).tolist()
    # Remove missing weights
    data1 = data1.dropna(subset=['value'], how='any')
    # Calculate val*weight
    data1['val*w'] = data1['value'] * data1['weight']
    # Calculate sum(val*weight)
    data1['sum(val*w)'] = data1.groupby(by_)['val*w'].transform('sum')
    # Calculate sum(weight)
    data1['sum(w)'] = data1.groupby(by_)['weight'].transform('sum')
    # Calculate sum(val*weight)/sum(weight)
    data1['wavg'] = data1['sum(val*w)'] / data1['sum(w)']
    # Drop duplicates
    data1 = data1.drop_duplicates(by_)
    # Keep columns
    data1 = data1[by_ + ['wavg']]
    # Decode variable names
    data1.columns = data1.columns.to_series().replace(decodedict).tolist()
    return data1


@njit
def wmean(data: np.ndarray, weights: np.ndarray) -> float:
    """Weighted average

    Parameters
    ----------
    data : numpy.ndarray
    weights : numpy.ndarray

    Returns
    -------
    float
    """
    assert data.ndim == 1, 'data is not 1-d array'
    assert weights.ndim == 1, 'weights is not 1-d array'
    assert data.shape[0] == weights.shape[0], \
        'data and weights are not of the same shape'
    numer = 0
    denom = 0
    for i in range(len(data)):
        if np.isfinite(data[i]) and np.isfinite(weights[i]):
            numer += data[i] * weights[i]
            denom += weights[i]
    if denom > 0:
        return numer / denom
    else:
        return np.nan


def des(data: Union[pd.Series, pd.DataFrame],
        dropna: bool=False,
        format: Union[None,str,list]=',.2f',
        columns: str='stats',
        html: bool=False,
        **kwargs):
    """Enhanced pandas describe() with NA counts and better formating

    Parameters
    ----------
    data : pd.Series or pd.DataFrame
        Input data
    dropna : bool, default: False
        If True, NA counts and percentages are displayed.
    format : None or str or list, default: None
        Format of statistics for each variable. If data is pd.Series, format
        must be either a str or a list of a single str.
    """
    assert isinstance(data, (pd.Series, pd.DataFrame)), 'data is not pandas.Series or pandas.DataFrame'
    assert columns in ['stats', 'vars'], 'Invalid value for columns'
    out = data.describe(**kwargs)
    # Change format
    if isinstance(data, pd.Series):
        if pd.api.types.is_list_like(format) and len(format) == 1:
            outstr_wocnt = out.drop(index='count').apply(
                lambda x: _format_to_str(x, format[0])
            )
        elif isinstance(format, str):
            outstr_wocnt = out.drop(index='count').apply(
                lambda x: _format_to_str(x, format)
            )
        else:
            raise TypeError('format is invalid for pandas.Series')
        outstr_cnt = pd.Series(
            [_format_to_str(out.loc['count'], ',.0f')], index=['count']
        )
    elif isinstance(data, pd.DataFrame):
        if isinstance(format, str):
            outstr_wocnt = out.drop(index='count').apply(
                lambda x: _format_to_str(x, format)
            )
        elif pd.api.types.is_list_like(format):
            assert len(format) == len(data.columns), \
                'format does not have the same length as data'
            outstr_wocnt = pd.concat([
                _format_to_str(out.drop(index='count').iloc[:, i], format[i])
                for i in range(len(out.columns))
            ], axis=1)
        else:
            raise TypeError('format is not list-like or a str')
        outstr_cnt = _format_to_str(out.loc['count'], ',.0f').to_frame().T
    else:
        raise TypeError('data should be pandas.Series or pandas.DataFrame')
    outstr = pd.concat([outstr_cnt, outstr_wocnt], axis=0)
    # Add NA counts
    if not dropna:
        na_count = _is_missing(data).sum(axis=0)
        na_pct = na_count / data.shape[0] * 100
        outstr.loc['#NA'] = _format_to_str(na_count, ',d')
        outstr.loc['%NA'] = _format_to_str(na_pct, '.2f', suffix='%')
    if isinstance(data, pd.Series):
        outstr = outstr.to_frame(data.name)
    if columns == 'stats':
        outstr = outstr.T
    # Output
    if html:
        return outstr.style.set_table_styles(
            [
                dict(selector='td', props=[('text-align', 'right')]),
                dict(selector='thead>tr>th', props=[('text-align', 'center')])
            ]
        )
    else:
        return outstr


@njit
def _float_to_int_numba(arr):
    out = np.ones_like(arr, 'int')
    for i in range(arr.shape[0]):
        out[i] = int(arr[i])
    return out


@njit
def _range_stat(datalist, func, interval, ncolout, *args):
    timearr = datalist[0]
    data = datalist[1]
    out = np.zeros((data.shape[0], ncolout))
    out.fill(np.nan)
    istart = 0
    iend = 1
    for i in range(data.shape[0]):
        while timearr[istart,0] < interval[0] + timearr[i,0]:
            istart += 1
        while iend < data.shape[0] and timearr[iend,0] <= interval[1] + timearr[i,0]:
            iend += 1
        out[i,:] = func([data[istart:iend,:]], *args)[0]
    return np.hstack((timearr, out))



def rangestat(
        data: pd.DataFrame,
        by: List[str],
        timevar: str,
        interval: Tuple[Union[int,float], Union[int,float]],
        func: callable,
        colargs: List[str],
        otherargs: tuple,
        colout: List[str]
) -> pd.DataFrame:

    assert ndup(data, by+[timevar]) == 0, \
        'data has duplicates over by and timevar'
    assert not data[timevar].isna().any(), 'Column timevar has NaN'
    assert isinstance(timevar, str), 'timevar is not a str'
    assert isinstance(interval, tuple), 'interval is not a tuple'
    assert len(interval) == 2, 'interval does not have size 2'
    for i, s in enumerate(interval):
        assert isinstance(s, (int, float)), \
            f'The {i+1}th element in interval is not a int or float'
    assert interval[0] <= interval[1], \
        f'interval[0] is not less than or equal to interval[1]'

    # Sort by by and timevar
    data.sort_values(by + [timevar])
    # Convert dtype (numba only supports list with homogeneous types)
    timedtype = data[timevar].to_numpy().dtype
    colargsdtype = data[colargs].to_numpy().dtype
    if timedtype != colargsdtype:
        if np.issubdtype(timedtype, np.integer) and \
            np.issubdtype(colargsdtype, np.floating):
            data[timevar] = data[timevar].astype('float')
    out = groupby_apply(
        data=data,
        by=by,
        func=_range_stat,
        colargs=[[timevar], colargs],
        otherargs=(func, interval, len(colout), *otherargs),
        colout=[timevar] + colout
    )
    out[timevar] = out[timevar].astype('int64')

    return out


def groupby_apply(
        data: pd.DataFrame,
        by: list,
        func: callable,
        colargs: List[list],
        otherargs: tuple,
        colout: list,
):
    """
    Fast groupby-apply using numba

    Parameters
    ----------
    data : pandas.DataFrame
        Input dataframe, should be sorted based on -by-.
    by : list of str
        Names of columns to group by
    func : callable numba function
        The numba function to be applied to each group. The first argument
        is a list (of length `len(colargs)`) of 2d ndarrays from columns
        specified in colargs. Other arguments are given via otherargs.
        The return value is also a 2d numpy array.
    colargs :  list of list of str
        Names of columns to be sent to the numba function. Then `len(colargs)` 2d
        ndarrays selected by these column names are sent to `func` as the
        first `len(colargs)` arguments.
    otherargs : tuple
        Other arguments (e.g. constant) to be sent to the numba function. Its
        order should match the arguments of the numba function
    colout : list of str
        Names of output columns

    Returns
    -------
    pd.DataFrame
    """
    # Check parameters
    if not isinstance(data, pd.DataFrame):
        raise TypeError('data is not a pandas.DataFrame')
    if not isinstance(by, list):
        raise TypeError('by is not a list')
    for x in by:
        if x not in data.columns:
            raise ValueError(f'{x} in "by" is not in columns of data')
    if data[by].isna().any(axis=None):
        raise ValueError('by-columns have NaN')
    if not isinstance(colargs, list):
        raise TypeError('colargs is not a list')
    for k, l in enumerate(colargs):
        if not isinstance(l, list):
            raise TypeError(f'The {k+1}th element in colargs is not a list')
        for i, c in enumerate(l):
            if not c in data.columns:
                raise ValueError(
                    f'The {i+1}th element ({c}) in the {k+1}th element '
                    f'in colargs is not in columns of data'
                )
    if not isinstance(otherargs, tuple):
        raise ValueError('otherargs is not a tuple')
    if not isinstance(colout, list):
        raise ValueError('colout is not a list')
    # Check if data is sorted on -by-
    byl = [list(row)[1:] for row in data[by].itertuples()]
    if not all(tuple(byl[i]) <= tuple(byl[i+1]) for i in range(data.shape[0]-1)):
        raise ValueError('data is not sorted on columns in "by"')
    # Match variable names
    by_name2id = {x: f'by{i}' for i, x in enumerate(by)}
    by_id2name = {f'by{i}': x for i, x in enumerate(by)}
    colargs_name2id = [
        {x: f'colargs{k}_{i}' for i, x in enumerate(colarg)}
        for k, colarg in enumerate(colargs)
    ]
    # New column names
    by_id = [f'by{i}' for i, x in enumerate(by)]
    colargs_id = [
        [f'colargs{k}_{i}' for i, x in enumerate(colarg)]
        for k, colarg in enumerate(colargs)
    ]
    # # Extract data
    # datatmp = data[by + colargs].sort_values(by)
    # # Rename
    # datatmp = datatmp.rename(columns=by_name2id).rename(columns=colargs_name2id)
    # Extract data and rename columns
    # (sort removed since it is not stable by default)
    datatmp = pd.concat([data[by].rename(columns=by_name2id)] + [
        data[colarg].rename(columns=colargs_name2id[k])
        for k, colarg in enumerate(colargs)
    ], axis=1)
    # ], axis=1).sort_values(by_id)
    # Get group id
    datatmp['grpid'] = datatmp.groupby(by_id).ngroup()
    # Create a link between by variables and group id
    grplk = datatmp[by_id + ['grpid']].drop_duplicates()
    # Apply the lower-level numba groupby function
    # NOTE: For pandas dataframe, selecting a single column returns a C-type
    # array, but selecting multiple columns return a Fortran-type array
    funcresnp = _groupby_apply_nb(
        datatmp['grpid'].to_numpy(),
        numba.typed.List([np.ascontiguousarray(datatmp[colarg_id].to_numpy())
         for colarg_id in colargs_id]),
        func,
        otherargs
    )
    funcres = pd.concat([
        pd.DataFrame(x, columns=['grpid'] + colout) for x in funcresnp
    ], axis=0).reset_index(drop=True)
    # funcres = pd.DataFrame(np.vstack(funcresnp), columns=['grpid'] + colout)
    # funcres['grpid'] = funcres['grpid'].astype('int')
    # Merge by-vars
    dataout = grplk.merge(funcres, on='grpid', how='left', validate='1:m')
    dataout = dataout.rename(columns=by_id2name)
    dataout = dataout.drop(columns=['grpid'])
    return dataout


@njit
def _groupby_apply_nb(
        grpids: np.ndarray,
        datalist: List[np.ndarray],
        func,
        otherargs
):
    """
    Groupby-Apply using numba given group id

    Parameters
    ----------
    grpids : 1d np.ndarray
        Group id
    datalist : list of 2d np.ndarray
        This list of arrays will be sent to func as the first argument
    func : callable
    otherargs: tuple

    Returns
    -------
    np.ndarray of type float
    """


    ngroups = int(grpids[-1])+1   # Number of groups
    nrows = grpids.shape[0]    # Number of rows
    reslist = []
    istart = 0
    for k in range(ngroups):
        # Find start and end rows of the group
        # (istart point to the start and iend-1 point to the end
        iend = istart + 1
        while iend < nrows and grpids[iend-1] == grpids[iend]:
            iend += 1
        res = func([data[istart:iend,:] for data in datalist], *otherargs)
        assert _is_2darray(res), 'The func must return a 2d ndarray'
        reslist.append(np.hstack((
            np.array([[k]]*res.shape[0]), res
        )))
        # Move to the next group
        istart = iend
    assert len(reslist) == ngroups
    return reslist



@generated_jit(nopython=True)
def _is_2darray(data):
    if isinstance(data, ARR_TYPES) and data.ndim == 2:
        return lambda data: True
    else:
        return lambda data: False



def _format_to_str(data: Union[float, pd.Series],
                   format: str,
                   prefix: str='',
                   suffix: str='') -> Union[str, pd.Series]:
    """Format data (series or number) into str.

    Parameters
    ----------
    data : float or pandas.Series
    format : str
    prefix : str, default: ''
    suffix : str, default: ''

    Returns
    -------
    str or pd.Series
    """
    formatfunc = lambda x: (prefix + '{:' + format + '}' + suffix).format(x)
    if isinstance(data, pd.Series):
        return data.apply(formatfunc)
    elif pd.api.types.is_scalar(data):
        return formatfunc(data)


import sqlite3
import pandas as pd
import numpy as np
from typing import Union

# SQL query
def dfsql(
        query: str,
        parse_dates: Union[list,None]=None,
        create_functions: Union[list,None]=None,
        create_aggregates: Union[list,None]=None,
        **kwargs
        ) -> pd.DataFrame:
    """Execute a SQL query

    Parameters
    ----------
    query : str
        The SQL query to be executed
    parse_dates : list of str or None
        The list of variables of datetime type
    create_functions : list or None
        Functions to be used in the SQL query
    create_aggregates : list or None
        Aggregate functions (classes) to be used in the SQL query
    **kwargs
        The key is the table name in the SQL query and the value is the
        corresponding pandas data frame.

    Returns
    -------
    pandas.DataFrame
        The output table from the SQL query
    """
    # Connect database
    conn = sqlite3.connect(':memory:')
    # Add tables to database
    for v in kwargs:
        kwargs[v].to_sql(v, conn, index=False)
    # Add functions to database
    if create_functions is not None:
        for f in create_functions:
            conn.create_function(*f)
    # Add aggregate to database
    if create_aggregates is not None:
        for f in create_aggregates:
            conn.create_aggregate(*f)
    # Enable tracing errors
    sqlite3.enable_callback_tracebacks(True)
    # Execute query
    out = pd.read_sql_query(query, conn, parse_dates=parse_dates)
    # Close database
    conn.close()
    # Convert datetime
    return out

class StdevFunc:
    """Standard deviation function to be used as an aggregate function in SQLite
    """

    def __init__(self):
        self.M = 0.0
        self.S = 0.0
        self.k = 0

    def step(self, value):
        try:
            # automatically convert text to float, like the rest of SQLite
            # if fails, skips this iteration, which also ignores nulls
            val = float(value)
            tM = self.M
            self.k += 1
            self.M += ((val - tM) / self.k)
            self.S += ((val - tM) * (val - self.M))
        except:
            pass

    def finalize(self):
        if self.k <= 1:  # avoid division by zero
            return np.nan
        else:
            return np.sqrt(self.S / (self.k-1))


class WeightedAverageFunc:
    """Weighted average function to be used as an aggregate function in SQLite
    """

    def __init__(self):
        self.numer = 0.0
        self.denom = 0.0

    def step(self, value, weight):
        try:
            # automatically convert text to float, like the rest of SQLite
            # if fails, skips this iteration, which also ignores nulls
            val = float(value)
            w = float(weight)
            self.denom += w
            self.numer += val * w
        except:
            pass

    def finalize(self):
        if self.denom == 0:  # avoid division by zero
            return np.nan
        else:
            return self.numer / self.denom



import pandas as pd


# Code ratings
mrtable = {
    'Aaa': 1,
    'Aa1': 2,
    'Aa2': 3,
    'Aa': 3,
    'Aa3': 4,
    'A1': 5,
    'A2': 6,
    'A': 6,
    'A3': 7,
    'Baa1': 8,
    'Baa2': 9,
    'Baa': 9,
    'Baa3': 10,
    'Ba1': 11,
    'Ba2': 12,
    'Ba': 12,
    'Ba3': 13,
    'B1': 14,
    'B2': 15,
    'B': 15,
    'B3': 16,
    'Caa1': 17,
    'Caa2': 18,
    'Caa': 18,
    'Caa3': 19,
    'Ca': 20,
    'C': 21,
    'D': 22
}
sprtable = {
    'AAA': 1,
    'AA+': 2,
    'AA': 3,
    'AA-': 4,
    'A+': 5,
    'A': 6,
    'A-': 7,
    'BBB+': 8,
    'BBB': 9,
    'BBB-': 10,
    'BB+': 11,
    'BB': 12,
    'BB-': 13,
    'B+': 14,
    'B': 15,
    'B-': 16,
    'CCC+': 17,
    'CCC': 18,
    'CCC-': 19,
    'CC': 20,
    'C': 21,
    'D': 22
}

def coderating(series: pd.Series, type: str) -> pd.Series:
    """Convert str credit rating to integer

    Parameters
    ----------
    series : pandas.Series
    type : str
        'mr' or 'spr'

    Returns
    -------
    pandas.Series
    """
    if type == 'mr':
        return series.apply(lambda x: mrtable.get(x, pd.NA)).astype('Int8')
    elif type == 'spr':
        return series.apply(lambda x: sprtable.get(x, pd.NA)).astype('Int8')
    else:
        raise ValueError('Invalid credit rating type!')



import pyempfin.datautils as pf
from pandas.testing import assert_frame_equal
from numpy.testing import assert_array_equal
import math
import numpy as np
import pandas as pd
import pytest
from numba import njit
import numba

@njit
def test():
    a = [np.ones((4,4)), np.zeros((4,4))]
    return [x[1:3,:] for x in a]
test()

class TestUtils:

    def test_des(self):
        df = pd.DataFrame({
            'x': [0, 1, np.nan, 3]
        })
        dftest = pf.des(df)
        assert dftest['%NA'].iloc[0] == '25.00%'

    def test_groupby_apply_nb(self):
        testarr = np.hstack((
            np.array([0]*15 + [1]*10)[:, np.newaxis],
            np.random.rand(25, 2)
        ))
        # Func that takes 1 argument and returns a number (wrapped in 2d ndarray)
        restest1 = pf._groupby_apply_nb(
            testarr[:,0], [testarr[:,1:2]], _mean, ()
        )
        restest1 = pd.DataFrame(
            [x[0] for x in restest1], columns=['grpid', 'mean']
        )
        rescorr1 = pd.DataFrame(testarr[:,:2], columns=['grpid', 'x']) \
            .groupby('grpid').apply(
            lambda x: _mean([x[['x']].to_numpy()])[0][0]
        )
        rescorr1 = pd.DataFrame(
            rescorr1.tolist(), index=rescorr1.index, columns=['mean']
        ).reset_index(drop=False)
        assert_frame_equal(restest1, rescorr1)
        # Func that takes 1 argument with additional arguments and
        # returns a row (wrapped in 2d ndarray)
        restest2 = pf._groupby_apply_nb(
            testarr[:,0], [testarr[:,1:]], _ols_njit, (10,)
        )
        restest2 = pd.DataFrame(
            [x[0] for x in restest2], columns=['grpid', 'coef0', 'coef1']
        )
        rescorr2 = pd.DataFrame(testarr, columns=['grpid', 'y', 'x'])\
            .groupby('grpid').apply(
            lambda x: _ols_njit([x[['y', 'x']].to_numpy()], 10)[0]
        )
        rescorr2 = pd.DataFrame(
            rescorr2.tolist(), index=rescorr2.index, columns=['coef0', 'coef1']
        ).reset_index(drop=False)
        assert_frame_equal(restest2, rescorr2)
        # Func that takes 2 argument (in a list) with additional arguments and
        # returns a row (wrapped in 2d ndarray)
        restest3 = pf._groupby_apply_nb(
            testarr[:,0], [testarr[:,1:2], testarr[:,2:3]], _ols2_njit, (10,)
        )
        restest3 = pd.DataFrame(
            [x[0] for x in restest3], columns=['grpid', 'coef0', 'coef1']
        )
        assert_frame_equal(restest3, rescorr2)
        # Func that returns higher dimensional ndarray


        @njit
        def _wrong_func(datalist):
            return np.random.rand(2, 2, 2)


        msg = 'The func must return a 2d ndarray'
        with pytest.raises(numba.core.errors.TypingError, match=msg):
            pf._groupby_apply_nb(
                testarr[:,0], [testarr[:,1:]], _wrong_func, ())




@njit
def _ols_njit(datalist, scaley):
    """Single-variable OLS regression with constant.
    The first column is y and the second column is x."""
    data = datalist[0]
    # Remove NA
    nnanmask = (~np.isnan(data[:, 0])) & (~np.isnan(data[:, 1]))
    ytmp = data[nnanmask, 0] * scaley
    xtmp = data[nnanmask, 1]
    N = ytmp.shape[0]
    if N >= 5:
        X = np.column_stack((np.ones(N), xtmp))
        XX = X.T @ X
        beta = np.linalg.inv(XX) @ X.T @ ytmp
        return np.array([list(beta)])
    else:
        return np.array([[np.nan, np.nan]])



@njit
def _ols2_njit(datalist, scaley):
    """Single-variable OLS regression with constant.
    The first column is y and the second column is x."""
    datay = datalist[0]
    datax = datalist[1]
    # Remove NA
    nnanmask = np.isfinite(datay[:,0]) & np.isfinite(datax[:,0])
    ytmp = datay[nnanmask,0] * scaley
    xtmp = datax[nnanmask,0]
    N = ytmp.shape[0]
    if N >= 5:
        X = np.column_stack((np.ones(N), xtmp))
        XX = X.T @ X
        beta = np.linalg.inv(XX) @ X.T @ ytmp
        return np.array([list(beta)])
    else:
        return np.array([[np.nan, np.nan]])


@njit
def _corr(datalist):
    data = datalist[0]
    nnanmask = np.isfinite(data[:,0]) & np.isfinite(data[:,1])
    if nnanmask.sum() > 0:
        m1 = data[nnanmask,0].mean()
        m2 = data[nnanmask,1].mean()
        std1 = data[nnanmask,0].std()
        std2 = data[nnanmask,1].std()
        if std1 > 0 and std2 > 0:
            return np.array([[
                (data[nnanmask,0]-m1).dot(data[nnanmask,1]-m2) / (
                        nnanmask.sum() * std1 * std2
                )
            ]])
        else:
            return np.array([[np.nan]])
    else:
        return np.array([[np.nan]])


@njit
def _corr2(datalist):
    data1 = datalist[0]
    data2 = datalist[1]
    nnanmask = np.isfinite(data1[:,0]) & np.isfinite(data2[:,0])
    if nnanmask.sum() > 0:
        m1 = data1[nnanmask,0].mean()
        m2 = data2[nnanmask,0].mean()
        std1 = data1[nnanmask,0].std()
        std2 = data2[nnanmask,0].std()
        if std1 > 0 and std2 > 0:
            return np.array([[
                (data1[nnanmask,0]-m1).dot(data2[nnanmask,0]-m2) / (
                        nnanmask.sum() * std1 * std2
                )
            ]])
        else:
            return np.array([[np.nan]])
    else:
        return np.array([[np.nan]])



@njit
def _mean(datalist):
    data = datalist[0]
    res = 0
    cnt = 0
    for i in range(data.size):
        if np.isfinite(data[i,0]):
            res += data[i,0]
            cnt += 1
    if cnt > 0:
        return np.array([[res / cnt]])
    else:
        return np.array([[np.nan]])

@njit
def _std(datalist):
    datam = _mean(datalist)[0][0]
    data = datalist[0]
    if np.isfinite(datam):
        res = 0
        cnt = 0
        for i in range(data.size):
            if np.isfinite(data[i,0]):
                res += (data[i,0] - datam) * (data[i,0] - datam)
                cnt += 1
        if cnt > 3:
            return np.array([[np.sqrt(res / cnt)]])
        else:
            return np.array([[np.nan]])
    else:
        return np.array([[np.nan]])

@pytest.fixture
def firm_bond_data():
    # Create the test data set
    nfirms = 2
    nbonds = 10
    ndates = 20
    droppct = 1/6
    napct = 1/3
    df = pd.DataFrame(
        {'e': np.random.randn(nfirms*nbonds*ndates),
         'x': np.random.rand(nfirms*nbonds*ndates)},
        pd.MultiIndex.from_product(
            [np.arange(nfirms), np.arange(nbonds), np.arange(ndates)],
            names=['firm', 'bond', 'date']
        )
    ).reset_index(drop=False)
    df['y'] = 5 + df['x'] * 10 + df['e']
    np.random.seed(0)
    # Randomly drop observations
    dropmask = np.random.choice(
        df.index, size=int(df.shape[0]*droppct), replace=False
    )
    df = df.drop(index=dropmask)
    # Randomly set x as NA
    xnamask = np.random.choice(
        df.index, size=int(df.shape[0]*napct), replace=False
    )
    df.loc[xnamask, 'x'] = np.nan
    # Randomly set y as NA
    ynamask = np.random.choice(
        df.index, size=int(df.shape[0]*napct), replace=False
    )
    df.loc[ynamask, 'y'] = np.nan
    return df


class TestAgg:

    def test_groupbyapply_simple(self):
        df = pd.DataFrame({
            'x': [0, 0, 0, 0, 1, 1, 1, 1, 1],
            'y': [2, 2, 3, 3, 2, 2, 2, 3, 3],
            'var1': [0, 1, 3, 4, 1, 2, 3, 4, 4],
        })

        @njit
        def testfunc(datalist):
            data = datalist[0]
            tmp = data[data[:,0] < data[0,1], 0]
            if tmp.shape[0] == 0:
                return np.array([[-999]])
            else:
                return np.array([[tmp.max()]])

        dfout = pf.groupby_apply(
            data=df,
            by=['x', 'y'],
            func=testfunc,
            colargs=[['var1', 'y']],
            otherargs=(),
            colout=['out']
        )

        dfcor = pd.DataFrame({
            'x': [0, 0, 1, 1],
            'y': [2, 3, 2, 3],
            'out': [1, -999, 1, -999]
        })

        assert_frame_equal(dfout, dfcor)

    def test_groupbyapply_simple_transform(self):
        N = 1000
        Ngx = 10
        Ngy = 20
        df = pd.DataFrame({
            'x': np.random.randint(0, Ngx, N),
            'y': np.random.randint(0, Ngy, N),
            'var1': np.random.rand(N),
        })
        df.sort_values(['x', 'y'], inplace=True)
        df.reset_index(drop=True, inplace=True)

        @njit
        def testfunc(datalist):
            """Return the same size as input"""
            return datalist[0]

        dfout = pf.groupby_apply(
            data=df,
            by=['x'],
            func=testfunc,
            colargs=[['var1']],
            otherargs=(),
            colout=['var1']
        )

        assert_frame_equal(df[['x', 'var1']], dfout)

    def test_groupbyapply(self, firm_bond_data):
        df = firm_bond_data
        # Check sorted
        msg = 'data is not sorted on columns in "by"'
        with pytest.raises(ValueError, match=msg):
            dftest1 = pf.groupby_apply(
                data=df, by=['firm', 'date'], func=_mean, colargs=[['x']],
                otherargs=(), colout=['mean']
            )
        # ---------- Func returning a single number (wrapped in a 2d ndarray)
        # Correct result
        df.sort_values(['firm', 'date'], inplace=True)
        dfcorr1 = df.groupby(['firm', 'date']).apply(
            lambda x: _mean([x[['x']].to_numpy()])[0][0]
        ).to_frame('mean').reset_index(drop=False)
        # Test result
        dftest1 = pf.groupby_apply(
            data=df, by=['firm', 'date'], func=_mean, colargs=[['x']],
            otherargs=(), colout=['mean']
        )
        assert_frame_equal(dfcorr1, dftest1)
        # ---------- Func with 1 argument (in list) and returning two numbers
        df.sort_values(['firm', 'date'], inplace=True)
        dfcorr2 = df.groupby(['firm', 'date']).apply(
            lambda x: _ols_njit([x[['y', 'x']].to_numpy()], 10)
        )
        dfcorr2 = pd.DataFrame(
            [x[0] for x in dfcorr2],
            index=dfcorr2.index,
            columns=['const', 'x']
        ).reset_index(drop=False)
        dftest2 = pf.groupby_apply(
            data=df, by=['firm', 'date'], func=_ols_njit, colargs=[['y', 'x']],
            otherargs=(10, ), colout=['const', 'x']
        )
        assert_frame_equal(dfcorr2, dftest2)
        # ---------- Func with 2 arguments (in list) and return two numbers
        df.sort_values(['firm', 'date'], inplace=True)
        dftest3 = pf.groupby_apply(
            data=df, by=['firm', 'date'], func=_ols2_njit, colargs=[['y'], ['x']],
            otherargs=(10, ), colout=['const', 'x']
        )
        assert_frame_equal(dfcorr2, dftest3)
        # ---------- Func with arguments and returning a 2d ndarray

        @njit
        def _test_func(datalist, nobs):
            data = datalist[0]
            if data.shape[0] >= nobs:
                return data[:nobs,:]
            else:
                return data

        # Correct result
        df.sort_values(['firm', 'date'], inplace=True)
        dfcorr4 = df.groupby(['firm', 'date']).apply(
            lambda x: _test_func([x[['y', 'x']].to_numpy()], 10)
        ).explode()
        dfcorr4 = pd.DataFrame(
            [x for x in dfcorr4],
            index=dfcorr4.index,
            columns=['y_', 'x_']
        ).reset_index(drop=False)
        # Test result
        df.sort_values(['firm', 'date'], inplace=True)
        dftest3 = pf.groupby_apply(
            data=df, by=['firm', 'date'], func=_test_func, colargs=[['y', 'x']],
            otherargs=(10, ), colout=['y_', 'x_']
        )
        assert_frame_equal(dfcorr4, dftest3)

    def test_rangestat(self, firm_bond_data):
        df = firm_bond_data

        lb = -12.5
        ub = 2.5
        # ---------- Prepare dates for correct answer
        dfsmpl = df[['firm', 'bond', 'date']].copy()
        smpldays = df[['date']].drop_duplicates()
        smpldays['smpl_date'] = smpldays['date'].apply(
            lambda x: np.arange(x+math.ceil(lb), x+math.floor(ub)+1)
        )
        smpldays= smpldays.explode('smpl_date')
        dfcorr = dfsmpl.merge(
            smpldays, on=['date'], how='left'
        )
        dfcorr = dfcorr.merge(
            df[['firm', 'bond', 'date', 'x', 'y']].rename(
                columns={'date': 'smpl_date'}
            ),
            on=['firm', 'bond', 'smpl_date'],
            how='inner'
        ).sort_values(['firm', 'bond', 'date', 'smpl_date'])
        # ---------- Func returning a single number (wrapped in a 2d ndarray)
        dfcorr1 = dfcorr.groupby(['firm', 'bond', 'date']).apply(
            lambda x: _std([x[['x']].to_numpy()])[0][0]
        ).to_frame('std').reset_index(drop=False)
        dftest1 = pf.rangestat(
            data=df, by=['firm', 'bond'], timevar='date', interval=(lb, ub),
            func=_std, colargs=['x'], otherargs=(), colout=['std']
        )
        assert_frame_equal(dfcorr1, dftest1)
        # ---------- Func with arguments and returning two numbers
        dfcorr2 = dfcorr.groupby(['firm', 'bond', 'date']).apply(
            lambda x: _ols_njit([x[['y', 'x']].to_numpy()], 10)
        )
        dfcorr2 = pd.DataFrame(
            [x[0] for x in dfcorr2],
            index=dfcorr2.index,
            columns=['const', 'x']
        ).reset_index(drop=False)
        dftest2 = pf.rangestat(
            data=df, by=['firm', 'bond'], timevar='date', interval=(lb, ub),
            func=_ols_njit, colargs=['y', 'x'], otherargs=(10,), colout=['const', 'x']
        )
        assert_frame_equal(dfcorr2, dftest2)


import datetime

class HtmlOutput:

    def __init__(self,
                 title: str='',
                 title_to_h1: bool=True,
                 insert_date: bool=True):
        """Create a new HTML file with specific title"""
        # HTML
        self._fullhtml = '<!DOCTYPE html>\n<html>\n<head>\n' + \
                         f'<meta charset="utf-8">\n<title>{title}</title>\n' + \
                         '</head>\n<body>\n</body>\n</html>'
        # Position of end-of-file
        self._endptr = self._fullhtml.find('</body>')
        # Add Heading 1
        if title_to_h1:
            self.insert_tag(title, 'h1')
        # Add date
        if insert_date:
            self.insert_str(
                'Date Created: ' +
                datetime.datetime.now().strftime('%Y-%m-%d (%a) %H:%M:%S') + '\n'
            )


    def _refresh_endptr(self):
        self._endptr = self._fullhtml.find('</body>')


    def insert_str(self, text: str):
        self._fullhtml = self._fullhtml[:self._endptr] + text + \
                         self._fullhtml[self._endptr:]
        self._refresh_endptr()


    def insert_tag(self, content: str, tag: str):
        self.insert_str(
            f'<{tag}>' + content + f'</{tag}>\n'
        )


    def save(self, filepath):
        with open(filepath, 'w') as f:
            f.write(self._fullhtml)


    def print(self):
        print(self._fullhtml)


    def __repr__(self):
        return self._fullhtml


    def __str__(self):
        return self._fullhtml


    @property
    def fullhtml(self):
        return self._fullhtml




from pyempfin.datetimeutils import *
import pandas as pd
import calendar
from pandas.testing import assert_series_equal

def test_yrdif():
    df1 = pd.DataFrame({
        'start': ['2010-08-27', '2010-08-27'],
        'end': ['2022-06-01', '2024-06-01']
    })
    df1['start'] = pd.to_datetime(df1['start'])
    df1['end'] = pd.to_datetime(df1['end'])
    df1['ttm'] = yrdif(df1['start'], df1['end'])

    def verify_func(start, end):
        res = 0
        if start.year != end.year:
            res += (pd.Timestamp(start.year, 12, 31).dayofyear - \
                start.dayofyear) / (366 if calendar.isleap(start.year) else 365)
            res += end.dayofyear / \
                (366 if calendar.isleap(end.year) else 365)
            res += end.year - start.year - 1
        else:
            res += (end.dayofyear - start.dayofyear) / \
               (366 if calendar.isleap(start.year) else 365)
        return res

    df1['ttm_corr'] = df1.apply(
        lambda x: verify_func(x['start'], x['end']),
        axis=1
    )

    assert_series_equal(df1['ttm'], df1['ttm_corr'], check_names=False)

def test_monthstart():
    df = pd.DataFrame({
        'test': [
            '1990-01-01', '1990-01-31', '2000-02-29', '2000-06-30', '2000-11-01',
            '2000-11-15'
        ],
        'correct': [
            '1990-01-01', '1990-01-01', '2000-02-01', '2000-06-01', '2000-11-01',
            '2000-11-01'
        ]
    })
    df['test'] = pd.to_datetime(df['test'])
    df['correct'] = pd.to_datetime(df['correct'])
    assert_series_equal(monthstart(df['test']), df['correct'], check_names=False)

def test_monthend():
    df = pd.DataFrame({
        'test': [
            '1990-01-01', '1990-01-31', '2000-02-29', '2000-06-30', '2000-11-01',
            '2000-11-15'
        ],
        'correct': [
            '1990-01-31', '1990-01-31', '2000-02-29', '2000-06-30', '2000-11-30',
            '2000-11-30'
        ]
    })
    df['test'] = pd.to_datetime(df['test'])
    df['correct'] = pd.to_datetime(df['correct'])
    assert_series_equal(monthend(df['test']), df['correct'], check_names=False)



import pandas as pd
import numpy as np
from joblib import Parallel, delayed
from typing import Union
from numba import njit
import datetime

def intck(freq: str, fdate: Union[pd.Series,pd.Timestamp], ldate: Union[pd.Series,pd.Timestamp]) -> Union[pd.Series,pd.Timestamp]:
    """Return the number of months or days between two dates, similar to SAS intck function
    
    Parameters
    ----------
    freq : str
        Allowed values are "month" and "day"
    fdate : pd.Series or pd.Timestamp
        Start date or a series of start dates.
    ldate : pd.Series or pd.Timestamp
        End date or a series of end dates.
    """    
    if isinstance(fdate, pd.Series) and isinstance(ldate, pd.Series):
        if freq == 'month':
            return (ldate.dt.year - fdate.dt.year) * 12 + (ldate.dt.month - fdate.dt.month)
        elif freq == 'day':
            return (ldate - fdate).dt.days

    elif isinstance(fdate, pd.Series) and isinstance(ldate, pd.Timestamp):
        if freq == 'month':
            return (ldate.year - fdate.dt.year) * 12 + (ldate.month - fdate.dt.month)
        elif freq == 'day':
            return (ldate - fdate).dt.days

    elif isinstance(fdate, pd.Timestamp) and isinstance(ldate, pd.Series):
        if freq == 'month':
            return (ldate.dt.year - fdate.year) * 12 + (ldate.dt.month - fdate.month)
        elif freq == 'day':
            return (ldate - fdate).dt.days
        
    elif isinstance(fdate, pd.Timestamp) and isinstance(ldate, pd.Timestamp):
        if freq == 'month':
            return (ldate.year - fdate.year) * 12 + (ldate.month - fdate.month)
        elif freq == 'day':
            return (ldate - fdate) / np.timedelta64(1, 'D')
        
    elif isinstance(fdate, datetime.date) and isinstance(ldate, datetime.date):
        if freq == 'month':
            return (ldate.year - fdate.year) * 12 + (ldate.month - fdate.month)
        elif freq == 'day':
            return (ldate - fdate) / datetime.timedelta(days=1)

def sas2date(d: pd.Series) -> pd.Series:
    """Convert a pandas series of SAS date to pandas datetime

    Parameters
    ----------
    d : pandas.Series
        The series containing SAS date in numeric form, i.e. the number of days
        from 1960-01-01. This is typically from a pandas dataframe that is read
        using the pyreadstat.read_sas7bdat function.
    
    Returns
    -------
    pandas.Series
        A pandas series of pandas datetime.
    """
    return pd.to_datetime('1960-01-01') + pd.to_timedelta(d, unit='D')

def date2sas(d: pd.Series) -> pd.Series:
    """Convert a pandas series of datetime to SAS date in int

    Parameters
    ----------
    d : pandas.Series
        Pandas datetime type.

    Returns
    -------
    pandas.Series
        Integer representing number of days from 1960-01-01.
    """
    return (d - pd.Timestamp('1960-01-01')).dt.days

def ymd2date(d: pd.Series) -> pd.Series:
    """Convert a pandas series of yyyymmdd date to pandas datetime

    Parameters
    ----------
    d : pandas.Series
        The series containing date in the form of 8-digit integer yyyymmdd.
    
    Returns
    -------
    pandas.Series
        A pandas series of pandas datetime.
    """
    return pd.to_datetime(d, format='%Y%m%d')

# Pandas datetime to YYYYMMDD
def date2ymd(d: Union[pd.Series,pd.Timestamp], type: str='int') -> Union[pd.Series,int,str]:
    """Convert a pandas series of pandas datetime into yyyymmdd format, either
    string or integer type.

    Parameters
    ----------
    d : pandas.Series or pandas.Timestamp
        A pandas series of pandas datatime or a single pandas.Timestamp
    type : str, default: 'int'
        Either 'int' or 'str' specifying the output type.

    Returns
    -------
    pandas.Series
        A pandas series of type str or int depending on the parameter `type` or
        a single str or int if the input is a pandas.Timestamp.
    """
    if isinstance(d, pd.Series) and pd.api.types.is_datetime64_ns_dtype(d):
        if type == 'str':
            return (d.dt.year * 10000 + d.dt.month * 100 + d.dt.day).astype(type).str[:8]
        else:
            return (d.dt.year * 10000 + d.dt.month * 100 + d.dt.day).astype(type)
    elif isinstance(d, pd.Timestamp):
        return pd.Series([d.year * 10000 + d.month * 100 + d.day]).astype(type).iloc[0]
    else:
        raise ValueError("Invalid argument: Neither a pandas Timestamp nor a Series of pandas Timestamp")
    
def monthstart(d: pd.Series) -> pd.Series:
    """Convert a pandas series of pandas datetime into the first dates of month

    Parameters
    ----------
    d : pandas.Series
        A pandas series of pandas datetime
    
    Returns
    -------
    pandas.Series
        A pandas series of pandas datetime containing the first dates of each
        month.
    """
    # return d.astype('datetime64[M]')
    return d + pd.offsets.MonthEnd(-1) + pd.offsets.Day(1)

# Return month end
def monthend(d: pd.Series) -> pd.Series:
    """Convert a pandas series of pandas datetime into the last dates of month

    Parameters
    ----------
    d : pandas.Series
        A pandas series of pandas datetime
    
    Returns
    -------
    pandas.Series
        A pandas series of pandas datetime containing the last dates of each
        month.
    """
    # return d.dt.to_period('M').dt.to_timestamp('M')
    return d + pd.tseries.offsets.MonthBegin() - pd.offsets.Day(1)

# def yrdif(start: Union[pd.Timestamp,datetime.date,pd.Series],
#           end: Union[pd.Timestamp,datetime.date,pd.Series],
#           basis: str='ACT/ACT') -> Union[float,np.array]:
#     """Mimic the yrdif function in SAS to calculate number of years between two
#     dates (assuming end-of-day).
#
#     Parameters
#     ----------
#     start : pandas.Timestamp, datetime.date, or a pandas.Series of them
#         The start date, i.e. the time 00:00.00 on the beginning of the date
#     end : pandas.Timestamp, datetime.date, or a pandas.Series of them
#         The end date, i.e. the time 00:00.00 on the beginning of the date
#     basis : str, default: 'ACT/ACT'
#         The day count convention to calculate the year fraction according to
#         the SAS function. Currently only the 'ACT/ACT' convention is
#         implemented.
#
#     Returns
#     -------
#     float or a pandas series of float
#         The number of years between the two days
#     """
#     if isinstance(start, pd.Timestamp):
#         start = np.datetime64(start)
#     if isinstance(end, pd.Timestamp):
#         end = np.datetime64(end)
#     if isinstance(start, np.datetime64) and isinstance(end, np.datetime64):
#         return _yrdif([start], [end], basis)[0]
#     elif isinstance(start, pd.Series) and isinstance(end, pd.Series) and start.shape[0] == end.shape[0]:
#         # TODO: Speed up this part using numba?
#         return pd.Series(
#             Parallel(n_jobs=4)(delayed(_yrdif)(
#                 start[i], end[i], basis) for i in range(start.shape[0])),
#             index=start.index)
#

@njit
def is_leap_year(data: int) -> bool:
    """Return if a year is a leap year"""
    return (np.mod(data, 400) == 0) | \
           ((np.mod(data, 4) == 0) & (np.mod(data, 100) != 0))


@njit
def _datdif_diff_month(start, end):
    """Return number of days between two dates in different months (assume end-of-day)"""
    yid = 0
    mid = 1
    did = 2
    dinm_noleap = [0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
    dinm_leap = [0, 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
    if is_leap_year(start[yid]):
        dinm = dinm_leap
    else:
        dinm = dinm_noleap
    outdays = 0
    # Days in the first month
    outdays += dinm[start[mid]] - start[did]
    # Days in the last month
    outdays += end[did]
    # Days in between
    for m in range(start[mid]+1, end[mid]):
        outdays += dinm[m]
    return outdays


@njit
def _datdif_same_year(start: Union[list,tuple,np.array], end: Union[list,tuple,np.array]):
    """Return number of days between two dates in the same year (assume end-of-day)"""
    yid = 0
    mid = 1
    did = 2
    if start[mid] == end[mid]:
        # The two days are in the same month
        return end[did] - start[did]
    elif start[mid] < end[mid]:
        return _datdif_diff_month(start, end)
    elif start[mid] > end[mid]:
        return -_datdif_diff_month(end, start)


@njit
def _yrdif_same_year(start: Union[list,tuple,np.array], end: Union[list,tuple,np.array]):
    """Return year fraction between two dates in the same year (assume end-of-day)"""
    yid = 0
    mid = 1
    did = 2
    ndays = _datdif_same_year(start, end)
    if is_leap_year(start[yid]):
        return ndays / 366
    else:
        return ndays / 365


@njit
def _datdif_diff_year(start, end):
    """Return number of days between two dates in different year (assume end-of-day)"""
    yid = 0
    outdays = 0
    # Days in the first year
    outdays += _datdif_same_year(start, (start[yid], 12, 31))
    # Days in the last year
    outdays += _datdif_same_year((end[yid], 1, 1), end) + 1
    # Days in between
    for y in range(start[yid] + 1, end[yid]):
        if is_leap_year(y):
            outdays += 366
        else:
            outdays += 365
    return outdays


@njit
def _yrdif_diff_year(start, end):
    """Return year fraction between two dates in different year (assume end-of-day)"""
    yid = 0
    yrfrac = 0
    # Days in the first year
    yrfrac += _datdif_same_year(start, (start[yid], 12, 31)) / \
              (366 if is_leap_year(start[yid]) else 365)
    # Days in the last year
    yrfrac += (_datdif_same_year((end[yid], 1, 1), end) + 1) / \
              (366 if is_leap_year(end[yid]) else 365)
    # Days in between
    yrfrac += end[yid] - start[yid] - 1
    return yrfrac


@njit
def _datdif(start: Union[list,tuple,np.array], end: Union[list,tuple,np.array]):
    """Return the number of days between two dates (assume end-of-day)"""
    yid = 0
    mid = 1
    did = 2
    if start[yid] == end[yid]:
        return _datdif_same_year(start, end)
    elif start[yid] < end[yid]:
        return _datdif_diff_year(start, end)
    else:
        return -_datdif_diff_year(end, start)


@njit
def _yrdif(start: Union[list,tuple,np.array], end: Union[list,tuple,np.array]):
    """Return the year fraction between two dates (assume end-of-day)"""
    yid = 0
    mid = 1
    did = 2
    if start[yid] == end[yid]:
        return _yrdif_same_year(start, end)
    elif start[yid] < end[yid]:
        return _yrdif_diff_year(start, end)
    elif start[yid] > end[yid]:
        return -_yrdif_diff_year(end, start)


@njit
def _datdif_arr(start: np.array, end: np.array):
    """Same as _datdif but input is an array of dates"""
    N = start.shape[0]
    out = np.zeros(N)
    out.fill(np.nan)
    for i in range(N):
        out[i] = _datdif(start[i,:], end[i,:])
    return out


@njit
def _yrdif_arr(start: np.array, end: np.array):
    """Same as _yrdif but input is an array of dates"""
    N = start.shape[0]
    out = np.zeros(N)
    out.fill(np.nan)
    for i in range(N):
        out[i] = _yrdif(start[i,:], end[i,:])
    return out


def datdif(
        start: Union[pd.Series,pd.Timestamp,datetime.date],
        end: Union[pd.Series,pd.Timestamp,datetime.date],
        basis: str='ACT/ACT'
):
    """Mimic the datdif function (ACT/ACT) in SAS to calculate number of days
    between two dates (assuming end-of-day).

    Parameters
    ----------
    start : pandas.Timestamp, datetime.date, or a pandas.Series of them
        The start date, i.e. the time 00:00.00 on the beginning of the date
    end : pandas.Timestamp, datetime.date, or a pandas.Series of them
        The end date, i.e. the time 00:00.00 on the beginning of the date
    basis : str, default: 'ACT/ACT'
        The day count convention to calculate the year fraction according to
        the SAS function. Currently only the 'ACT/ACT' convention is
        implemented.

    Returns
    -------
    float or a pandas series of float
        The number of years between the two days
    """
    singledatetypes = (pd.Timestamp, datetime.date)
    if isinstance(start, pd.Series) or isinstance(end, pd.Series):
        if isinstance(start, pd.Series) and isinstance(end, pd.Series):
            assert start.shape[0] == end.shape[0], 'Start and End has different lengths.'
            startin = np.vstack((start.dt.year, start.dt.month, start.dt.day)).T
            endin = np.vstack((end.dt.year, end.dt.month, end.dt.day)).T
        elif isinstance(start, pd.Series) and isinstance(end, singledatetypes):
            startin = np.vstack((start.dt.year, start.dt.month, start.dt.day)).T
            endin = np.tile([end.year, end.month, end.day], (start.shape[0], 1))
        elif isinstance(start, singledatetypes) and isinstance(end, pd.Series):
            startin = np.tile([start.year, start.month, start.day], (end.shape[0], 1))
            endin = np.vstack((end.dt.year, end.dt.month, end.dt.day)).T
        out = _datdif_arr(startin, endin)
    else:
        startin = (start.year, start.month, start.day)
        endin = (end.year, end.month, end.day)
        out = _datdif(startin, endin)
    return out


def yrdif(
        start: Union[pd.Series,pd.Timestamp,datetime.date],
        end: Union[pd.Series,pd.Timestamp,datetime.date],
        basis: str = 'ACT/ACT'
):
    """Mimic the yrdif function (ACT/ACT) in SAS to calculate number of years
    between two dates (assuming end-of-day). If the two days are in the same
    year, then the total number of days in the year is used as the denominator
    (366 for leap year and 365 for non-leap year). If the two days are in
    different years, then the year fraction is calculated year by year. For
    the first year, the denominator is the total number of days in the first
    year. For the last year, the denominator is the total number of days in the
    last year. For each year in between, the year fraction is incremented by 1.

    There is no requirement that the start date must be before the end date. If
    the start date is after the end date, the result is a negative number whose
    absolute value is the result if the start and end date are swapped.

    Parameters
    ----------
    start : pandas.Timestamp, datetime.date, or a pandas.Series of them
        The start date, i.e. the time 00:00.00 on the beginning of the date
    end : pandas.Timestamp, datetime.date, or a pandas.Series of them
        The end date, i.e. the time 00:00.00 on the beginning of the date
    basis : str, default: 'ACT/ACT'
        The day count convention to calculate the year fraction according to
        the SAS function. Currently only the 'ACT/ACT' convention is
        implemented.

    Returns
    -------
    float or a pandas series of float
        The number of years between the two days
    """
    singledatetypes = (pd.Timestamp, datetime.date)
    if isinstance(start, pd.Series) or isinstance(end, pd.Series):
        if isinstance(start, pd.Series) and isinstance(end, pd.Series):
            assert start.shape[0] == end.shape[0], 'Start and End has different lengths.'
            startin = np.vstack((start.dt.year, start.dt.month, start.dt.day)).T
            endin = np.vstack((end.dt.year, end.dt.month, end.dt.day)).T
        elif isinstance(start, pd.Series) and isinstance(end, singledatetypes):
            startin = np.vstack((start.dt.year, start.dt.month, start.dt.day)).T
            endin = np.tile([end.year, end.month, end.day], (start.shape[0], 1))
        elif isinstance(start, singledatetypes) and isinstance(end, pd.Series):
            startin = np.tile([start.year, start.month, start.day], (end.shape[0], 1))
            endin = np.vstack((end.dt.year, end.dt.month, end.dt.day)).T
        out = _yrdif_arr(startin, endin)
    else:
        startin = (start.year, start.month, start.day)
        endin = (end.year, end.month, end.day)
        out = _yrdif(startin, endin)
    return out





import pytest
import numpy as np
import pandas as pd
from pyempfin.dfutils import keepna, dropvar, rename

@pytest.fixture
def df():
    return pd.DataFrame(np.random.rand(10, 3), columns=['a', 'b', 'c'])


def test_rename(df):
    df1 = rename(df, a='A', b='B')
    assert df1.columns.tolist() == ['A', 'B', 'c']

def test_dropvar(df):
    df1 = dropvar(df, 'a b')
    assert df1.columns.tolist() == ['c']
    with pytest.raises(AssertionError):
        df2 = dropvar(df, 'ab')


def test_keepna(df):
    df = pd.DataFrame(np.random.rand(10, 3), columns=['a','b','c'])
    df.loc[3, 'a'] = np.nan
    df.loc[4, 'a'] = np.nan
    df.loc[4, 'b'] = np.nan
    df.loc[5, 'a'] = np.nan
    df.loc[5, 'b'] = np.nan
    df.loc[5, 'c'] = np.nan

    df1 = keepna(df, subset=['a'], how='any')
    assert df1.index.tolist() == [3, 4, 5]

    df2 = keepna(df, subset=['a', 'b'], how='any')
    assert df2.index.tolist() == [3, 4, 5]

    df3 = keepna(df, subset=['a', 'b'], how='all')
    assert df3.index.tolist() == [4, 5]

    df4 = keepna(df, subset=['a', 'b', 'c'], how='any')
    assert df4.index.tolist() == [3, 4, 5]

    df5 = keepna(df, subset=['a', 'b', 'c'], how='all')
    assert df5.index.tolist() == [5]


import pandas as pd
import numpy as np
import sqlite3
from numba import njit
from joblib import Parallel, delayed
from functools import partial

from .datetimeutils import *
from .sqlutils import *
from .ffi import *
from .logger import *
from .xsap import *
from .datautils import *
from .dfutils import *
from .htmloutput import *





import numpy as np
import pandas as pd
from numba import njit
from joblib import Parallel, delayed, cpu_count
from scipy.stats import t as tdist
from typing import Union, Optional, Tuple, List, Callable
import re
from functools import partial
from statsmodels.regression.linear_model import OLS, OLSResults
from statsmodels.stats.sandwich_covariance import cov_hac
import scipy.stats
from numba import njit, int32, float64
import warnings
from .datautils import groupby_apply


# =============================================================================#
# Functions to estimate beta
# =============================================================================#

def estbeta(leftdata: pd.DataFrame, rightdata: pd.DataFrame, models: dict,
            window: tuple, minobs: int, hasconst: bool, ncore: int=cpu_count()
            ) -> pd.DataFrame:
    """Estimate beta given a list of models. leftdata is a pandas.DataFrame of
    dependent variables (returns) with MultiIndex representing assets and periods.
    rightdata is a pandas.DataFrame with index representing periods and columns
    of independent variables (factors) to estimate beta.

    Parameters:
    ----------
    leftdata : pandas.DataFrame
        The index should be a MultiIndex whose level 0 represents assets and
        level 1 represents time periods
    rightdata : pandas.DataFrame
        The index should be a MultiIndex whose level 0 represents assets and
        level 1 represents time periods. This DataFrame will be reindexed
        according to `leftdata`
    model : dict (keys: str, values: list of str)
        The keys are model names (used for referencing and output) and the
        corresponding values are lists of names of variables in the model. The
        first element represents the dependent variable and should be in the
        columns of `leftdata`. The rest represents independent variables and
        should be in the columns of `rightdata`.
    window : tuple of 2 integers
        The period used to estimate beta. For example, `(-24,-1)` means using
        data between `t-24` and `t-1` to estimate beta at time `t`.
    minobs : int
        Minimum number of observations to estimate beta
    hasconst : bool
        If True, a constant term is added to the regression.
    ncore : int, default `cpu_count()`
        The number of cores used for parallel computing. The default value is
        the total number of cores in the computer.

    Returns
    ----------
    output : DataFrame
        Its index is the same as that of `leftdata`. Its column is also a
        MultiIndex, whose number of levels depending on whether the columns
        of `rightdata` is a MultiIndex. Its level 0 represents different models
        (keys in `models`). Its other levels represent the betas for each
        factors (values in `models`) with the same column names as the factor.
        Hence, if the columns of `rightdata` is not a MultiIndex, then the columns
        of the returned dataframe have only 2 levels. If the columns of `rightdata`
        is a MultiIndex with 2 levels, then the columns of the returned dataframe
        have 3 levels.
    """
    outbeta = pd.concat([
        estbeta1m(
            leftdata=leftdata[x[0]],
            rightdata=rightdata[x[1:]],
            model=x[1:],
            window=window,
            minobs=minobs,
            hasconst=hasconst,
            ncore=ncore
        ) for x in models.values()
    ], axis=1, keys=models.keys(), join='outer'
    )
    return outbeta

def estbeta1m(leftdata: pd.Series, rightdata: pd.DataFrame, model: list,
              window: tuple, minobs: int, hasconst: bool, ncore: int=cpu_count()
              ) -> pd.DataFrame:
    """Estimate beta for a single model. leftdata is a pandas.Series of
    dependent variable (returns) with MultiIndex representing assets and periods.
    rightdata is a pandas.DataFrame with index representing periods and columns
    of independent variables (factors) to estimate beta.

    Parameters
    ----------
    leftdata : pandas.Series
        The index should be a MultiIndex whose level 0 represents assets and
        level 1 represents time periods
    rightdata : pandas.DataFrame
        The index represents time periods and is the same as the level 1 of
        the index of `leftdata`.
    model : list
        The elements should be a subset of column names of rightdata.
    window : tuple of 2 integers
        The period used to estimate beta. For example, `(-24,-1)` means using
        data between `t-24` and `t-1` (both inclusive) to estimate beta at
        time `t`. The first element cannot be greater than the second one
    minobs : int
        Minimum number of observations to estimate beta
    hasconst : bool
        If True, a constant term is added to the regression.
    ncore: int
        Number of cores for parallel computing

    Returns
    -------
    pandas.DataFrame
        The index is the same as `leftdata`. Each column represents the estimated
        beta.
    """
    # Check arguments
    if not isinstance(leftdata, pd.Series):
        raise TypeError('leftdata is not a pandas.Series')
    if not isinstance(rightdata, pd.DataFrame):
        raise TypeError('rightdata is not a pandas.DataFrame')
    if not set(model).issubset(set(rightdata.columns)):
        raise ValueError('elements of model is not a subset of columns of rightdata')
    if not len(window) == 2:
        raise ValueError('window does not have length 2')
    if not window[0] <= window[1]:
        raise ValueError('start time is after end time for window')
    # Drop NaN and construct matrix for return
    leftmat = leftdata.dropna().unstack(level=0).sort_index()
    rightdata = rightdata.dropna(how='all').reindex(leftmat.index)
    # Construct X
    exog = rightdata[model].to_numpy()
    out = pd.DataFrame(
        np.concatenate(
            Parallel(n_jobs=ncore)(
                delayed(
                    lambda x: _get_beta_njit(
                        x, exog, window, minobs, hasconst)
                )(leftmat[v].to_numpy())
                for v in leftmat.columns), axis=1
        ),
        index=leftmat.index,
        columns=pd.MultiIndex.from_product([leftmat.columns, np.arange(len(model))])
    )
    # If model is a list of tuples (so rightdata has MultiIndex column)
    out = out.stack(level=0)
    if isinstance(rightdata.columns, pd.MultiIndex):
        out = out.set_axis(pd.MultiIndex.from_tuples(model), axis='columns')
    else:
        out = out.set_axis(model, axis='columns')
    out = out.swaplevel(i=1, j=0, axis='index') \
        .sort_index()
    return out



@njit
def _get_beta_njit(endogmat: np.ndarray, exogmat: np.ndarray, window: tuple,
                   minobs: int, hasconst: bool
                   ) -> np.ndarray:
    """Calculate rolling-window beta

    Parameters
    ----------
    endogmat : 1-D numpy.ndarray
        A vector of returns indexed by time.
    exogmat : 2-D numpy.ndarray
        T-by-K matrix, Rows represent time periods and columns represent
        independent variables.
    window : tuple of 2 integers
        The period used to estimate beta. For example, `(-24,-1)` means using
        data between `t-24` and `t-1` to estimate beta at time `t`.
    minobs : int
        Minimum number of observations to estimate beta
    hasconst : bool
        If True, a constant term is added to the regression.

    Returns
    -------
    2-D numpy.ndarray
        T-by-K matrix. Rows represent time periods. Each column
        represent beta for each factor. The intercept is not returned.
    """
    return _get_rolling_stats_njit(
        endogmat, exogmat, window, minobs,
        _estimate_beta_njit, (hasconst,), exogmat.shape[1],
    )


@njit
def _get_rolling_stats_njit(
        endogmat: np.ndarray, exogmat: np.ndarray, window: tuple, minobs: int,
        func: Callable, func_args: tuple, func_retsize: int) -> np.ndarray:
    """Calculate rolling-window statistics

    Parameters
    ----------
    endogmat : 1-D numpy.ndarray
        A vector of returns indexed by time.
    exogmat : 2-D numpy.ndarray
        T-by-K matrix, Rows represent time periods and columns represent
        independent variables.
    window : tuple of 2 integers
        The period used to calculate statistics. For example, `(-24,-1)` means using
        data between `t-24` and `t-1` to calculate statistics at time `t`.
    minobs : int
        Minimum number of observations to calculate statistics
    func : Callable
        Function to be applied for each period to calculate statistics
        in the rolling window. The function has signature (endogmat,exogmat,*args).
        The function returns a 1D vector with size func_retsize
    func_args : tuple
        Other arguments to be passed to func.
    func_retsize : int
        Size of the returned vector

    Returns
    -------
    2-D numpy.ndarray
        T-by-K matrix. Rows represent time periods. Each column
        represent beta for each factor. The intercept is not returned.
    """
    # Number of periods, variables,
    nper = endogmat.shape[0]
    # Initialization
    outstats = np.zeros((nper, func_retsize))
    outstats.fill(np.nan)
    # Loop
    for t in range(-window[0], nper):
        # Note that right boundary is not included in array slicing
        endogtmp = endogmat[t+window[0]:t+window[1]+1]
        exogtmp = exogmat[t+window[0]:t+window[1]+1, :]
        # Generate indicator
        nnanmask = np.isfinite(endogtmp)
        for j in range(window[1]-window[0]+1):
            nnanmask[j] = np.isfinite(exogtmp[j, :]).all() * nnanmask[j]
        X = exogtmp[nnanmask, :]
        Y = endogtmp[nnanmask]
        XX = X.T @ X
        if (len(Y) >= minobs) and (np.linalg.matrix_rank(XX) == X.shape[1]):
            outstats[t, :] = func(Y, X, *func_args)
            # if hasconst:
            #     outstats[t, :] = (np.linalg.inv(XX) @ X.T @ Y)[:-1]
            # else:
            #     outstats[t, :] = np.linalg.inv(XX) @ X.T @ Y
    return outstats


@njit
def _estimate_beta_njit(endogNotNull, exogNotNull, hasconst):
    X = exogNotNull
    if hasconst:
        X = np.hstack((X, np.ones((X.shape[0], 1))))
    Y = endogNotNull
    XX = X.T @ X
    if hasconst:
        return (np.linalg.inv(XX) @ X.T @ Y)[:-1]
    else:
        return np.linalg.inv(XX) @ X.T @ Y


# =============================================================================#
# Fama-MacBeth Regression
# =============================================================================#

def fmreg(leftdata: pd.DataFrame, rightdata: pd.DataFrame, models: list,
        maxlag: int, roworder: list, hasconst: bool, scale: float,
        getlambda: bool, winsorcuts: Union[tuple,None]=None,
        winsorindeponly: bool=True, estfmt: tuple=('.3f', '.2f')
          ) -> Union[pd.DataFrame, Tuple[pd.DataFrame, pd.Series]]:
    """Estimate FM regression for one model specification.

    Parameters
    ----------
    leftdata : pd.DataFrame
        Dataframe containing dependent variables. Its index is a MultiIndex,
        whose level 0 represents entities and its level 1 represents time
        periods.
    rightdata : pd.DataFrame
        Dataframe containing independent variables. Its index is a MultiIndex,
        whose level 0 represents entities and its level 1 represents time
        periods. The columns can be either a index or a MultiIndex with 2 levels.
    models : list of list
        A list of models. Each element represents a model containing names.
        For each model, the elements can be used to select columns in
        leftdata and rightdata, even when their columns are MultiIndex. The
        first element represents the dependent variable and the others
        represent independent variables
    maxlag : int
        Maximum number of lags for Newey-West estimator
    roworder : list of str
        The order of rows in the output. If the columns of `rightdata` is a
        MultiIndex with 2 levels, then `roworder` should match level 1 of the
        MultiIndex.
    hasconst : bool
        If True, a constant term is added in the model.
    scale : float
        The estimated coefficients will be multipled by this number.
    winsorcuts : tuple or None
        A tuple of lower and upper percentiles to winsorize variables in each
        cross-sectional regression. Both should be between 0 and 1 and the
        lower one should be less than the upper one.
    winsorindeponly : bool
        If True, only independent variables are winsorized.
    estfmt : tuple
        A 2-D tuple of str specifying the format of the coefficient and tstat,
        respectively.
    getlambda : bool
        If True,
    Return
    ------
    pd.DataFrame
        Formatted regression table
    pd.DataFrame, pd.Series
        If getlambda=True, then also return estimated lambda in a pandas.Series
    """
    # Check arguments
    if not isinstance(leftdata, pd.DataFrame):
        raise TypeError('leftdata is not a pandas.DataFrame')
    if not isinstance(rightdata, pd.DataFrame):
        raise TypeError('rightdata is not a pandas.DataFrame')
    if isinstance(leftdata.columns, pd.MultiIndex):
        if not leftdata.columns.nlevels == 2:
            raise ValueError('if leftdata has a MultiIndex column names, '
                             'its level must be 2')
    else:
        if not isinstance(leftdata.columns, pd.Index):
            raise ValueError('if leftdata does not have a MultiIndex column '
                             'names, it must be a pandas.Index')
    if isinstance(rightdata.columns, pd.MultiIndex):
        if not rightdata.columns.nlevels == 2:
            raise ValueError('if rightdata has a MultiIndex column names, '
                             'its level must be 2')
        if not set(roworder).issubset(set(rightdata.columns.levels[1])):
            raise ValueError('roworder must be a subset of level 1 of columns '
                             'of rightdata if its column names is a MultiIndex')
    else:
        if not isinstance(rightdata.columns, pd.Index):
            raise ValueError('if rightdata does not have a MultiIndex column '
                             'names, it must be a pandas.Index')
        if not set(roworder).issubset(set(rightdata.columns)):
            raise ValueError('roworder must be a subset of columns of rightdata')
    if not isinstance(hasconst, bool):
        raise ValueError('hasconst must be a boolean')
    if winsorcuts is not None:
        if not isinstance(winsorcuts, tuple):
            raise ValueError('winsorcuts must be a tuple or None')
        if not (len(winsorcuts) == 2):
            raise ValueError('winsorcuts must be a tuple of length 2')
        if not (0<=winsorcuts[0]<=1):
            raise ValueError('winsorcuts[0] must be between 0 and 1')
        if not (0<=winsorcuts[1]<=1):
            raise ValueError('winsorcuts[1] must be between 0 and 1')
        if not (winsorcuts[0]<winsorcuts[1]):
            raise ValueError('winsorcuts[0] must be less than winsorcuts[1]')
    if not isinstance(winsorindeponly, bool):
        raise ValueError('winsorindeponly must be a boolean')
    if not (isinstance(estfmt, tuple) and len(estfmt) == 2):
        raise ValueError('estfmt must be a tuple with length 2')
    for s in estfmt:
        if not isinstance(s, str):
            raise ValueError(f'{s} in estfmt is not a str')

    estout = FMResult()
    for model in models:
        # Check arguments
        if not (isinstance(model, list) and len(model) >= 2):
            raise ValueError(f'The model ({model}) is not a list of length 2')
        if model[0] not in leftdata.columns:
            raise ValueError(f'model[0] of the model ({model}) not in columns'
                             f' of leftdata')
        if not set(model[1:]).issubset(set(rightdata.columns.to_flat_index())):
            raise ValueError(f'model[1:] of the model ({model}) not in columns'
                             f' of rightdata')
        # ---------- Prepare LHS data set
        if isinstance(leftdata.columns, pd.MultiIndex):
            # Note that model[0] is not a list, so double brackets ensure that a
            # dataframe is returned
            leftdatatmp = leftdata[[model[0]]]
            depvar = leftdatatmp.columns.get_level_values(1)[0]
            leftdatatmp.columns = [depvar]
        else:
            depvar = model[0]
            leftdatatmp = leftdata[[model[0]]]
            # No need to assign leftdata.columns here
        # ---------- Prepare RHS data set
        rightdatatmp = rightdata.reindex(leftdata.index)[model[1:]]
        if isinstance(rightdata.columns, pd.MultiIndex):
            # Note that model[1:] is a list, so single brackets are enough to
            # return a dataframe
            # This is a MultiIndex
            indepvars = list(rightdatatmp.columns.get_level_values(1))
            rightdatatmp.columns = indepvars
        else:
            indepvars = model[1:]
            # No need to assign rightdata.columns here
        # Add Constant
        if hasconst:
            indepvars += ['Constant']
        # Get entity and time variable names
        indvar = leftdatatmp.index.levels[0].name
        timevar = leftdatatmp.index.levels[1].name
        # ----------- Convert dataframe into matrix form
        # Get matrix form of dependent variable
        endogmat = leftdatatmp.reset_index().pivot(
            index=timevar,
            columns=indvar,
            values=leftdatatmp.columns[0]
        ).to_numpy()
        # Get matrix form of independent variable
        exogmat = np.array([rightdatatmp[[v]].reset_index().pivot(
            index=timevar,
            columns=indvar,
            values=v
        ) for v in rightdatatmp.columns])
        # Swap axis
        exogmat = exogmat.swapaxes(0, 1)
        # ----------- Fama-MacBeth Regression
        # Get estimates
        coef, stderr, nobs, nobsavg, nper, ar2 = _fmreg_njit(
            endogmat,
            exogmat,
            maxlag,
            hasconst,
            winsorcuts,
            winsorindeponly
        )
        # ----------- Format output
        estout.add(coef, stderr, indepvars, nper-1,
                   stat={
                       'N': nobs,
                       'Periods': nper,
                       'N/period': nobsavg,
                       'Adjusted $R^2$': ar2
                   })
    estouttab = estout.output(
        type='tstat',
        estfmt=estfmt,
        statfmt=(',.0f', ',.0f', ',.1f', '.3f'),
        roworder=roworder + ['Constant'] if hasconst else roworder,
        scale=scale
    )
    if not getlambda:
        return estouttab
    else:
        outlambda = pd.concat([
            pd.Series({v.label: v.param for v in eststo['est']})
            for eststo in estout.eststo
        ], axis=1)
        outlambda.reindex(roworder + ['Constant'] if hasconst else roworder)
        return estouttab, outlambda


@njit
def _fmreg_njit(endogmat, exogmat, maxlag, hasconst, winsorcuts, winsorindeponly):
    # Verify parameters
    assert endogmat.ndim == 2, 'The number of dimensions of endog is not 2'
    assert exogmat.ndim == 3, 'The number of dimensions of endog is not 3'
    # Degress of freedom
    if hasconst:
        K = exogmat.shape[1] + 1
    else:
        K = exogmat.shape[1]
    # Total number of periods (including periods with missing variables)
    nper = exogmat.shape[0]
    # ----------- First stage
    # Lambdas
    lambdas = np.zeros((nper, K))
    lambdas.fill(np.nan)
    # Adjusted R2
    ar2 = np.zeros(nper)
    ar2.fill(np.nan)
    # Number of observations
    nobs = np.zeros(nper)
    nobs.fill(np.nan)
    # Loop over each row
    for t in range(nper):
        exogtmp = exogmat[t, :].T
        endogtmp = endogmat[t, :].T
        # Get nnanmask
        nnanmask = np.isfinite(endogtmp)
        for i in range(endogtmp.shape[0]):
            nnanmask[i] = nnanmask[i] * np.isfinite(exogtmp[i, :]).all()
        if hasconst:
            X = np.hstack((exogtmp, np.ones((endogtmp.shape[0], 1))))
        else:
            X = exogtmp
        X = X[nnanmask, :]
        Y = endogtmp[nnanmask]
        if X.size > 0 and Y.size > 0:
            # Winsor variables
            if winsorcuts is not None:
                for i in range(K):
                    X[:, i] = _winsor_njit(X[:, i], winsorcuts, 'inner')
                if not winsorindeponly:
                    Y = _winsor_njit(Y, winsorcuts, 'inner')
            XX = X.T @ X
            if X.shape[0] > X.shape[1] and np.linalg.matrix_rank(XX) == X.shape[1]:
                beta = np.linalg.inv(XX) @ X.T @ Y
                resid = Y - X @ beta
                tss = (Y - Y.mean()).dot(Y - Y.mean())
                rss = resid.dot(resid)
                ar2[t] = 1 - rss / tss * (Y.shape[0] - 1*hasconst) / (Y.shape[0] - K)
                lambdas[t, :] = beta.T
                nobs[t] = Y.shape[0]
    # ----------- Second stage
    # Calculate mean
    coef = np.zeros(K)
    for i in range(K):
        coef[i] = np.nanmean(lambdas[:, i])
    # Calculate standard error
    stderr = np.zeros(K)
    for i in range(K):
        stderr[i] = _newey_njit(lambdas[:, i]-coef[i], maxlag)
    # Calculate average adjusted R2
    adjr2 = np.nanmean(ar2)
    # Calculate total number of observations
    N = np.nansum(nobs)
    # Calculate average number of observations per period
    Navg = np.nanmean(nobs)
    # Number of periods (with value estimates)
    Nperout = np.isfinite(ar2).sum()
    return coef, stderr, N, Navg, Nperout, adjr2


@njit
def _newey_njit(xe: np.ndarray, maxlag: int):
    """Calculate Newey-west estimator of standard error given demeaned series

    Parameters
    ----------
    xe: pd.np.ndarray
        Demeaned series of float
    maxlag: int
        Maximum lag in Newey-West estimator
    Return
    ------
    float
        Newey-West estimator of standard error
    """
    # Remove NaN
    xe = xe[np.isfinite(xe)]
    T = xe.shape[0]
    if T <= maxlag:
        return np.nan
    # Calculate covariances
    cov = np.zeros(maxlag+1)
    cov[0] = xe.dot(xe) / T
    for i in range(1, maxlag+1):
        cov[i] = xe[:-i].dot(xe[i:]) / T
    # The estimator (Positive and negative parts are combined)
    out = cov[0] + 2 * (1-np.arange(1, maxlag+1)/(maxlag+1)).dot(cov[1:])
    return np.sqrt(out / (T-1))


@njit
def _winsor_njit(data: np.ndarray, cuts: tuple, interpolation: str):
    # Select non-missing values
    datatmp = data[np.isfinite(data)]
    N = datatmp.shape[0]
    datatmp.sort()
    # Get quantile
    if interpolation == 'inner':
        lowerbound = datatmp[int(np.ceil((N-1)*cuts[0]))]
        upperbound = datatmp[int(np.floor((N-1)*cuts[1]))]
    elif interpolation == 'outer':
        lowerbound = datatmp[int(np.floor((N-1)*cuts[0]))]
        upperbound = datatmp[int(np.ceil((N-1)*cuts[1]))]
    else:
        raise ValueError('Invalid interpolation type.')
    out = np.zeros(data.shape)
    for i in range(data.shape[0]):
        if np.isfinite(data[i]) and data[i] < lowerbound:
            out[i] = lowerbound
        elif np.isfinite(data[i]) and data[i] > upperbound:
            out[i] = upperbound
        else:
            out[i] = data[i]
    return out


def tscssum(data: pd.DataFrame, by: list, subset: Union[list,None]=None,
            percentiles: tuple=(.01, .05, .50, .95, .99)) -> pd.DataFrame:
    """Print time-series average of cross-sectional summary statistics for numeric columns

    Parameters
    ----------
    data : pandas.DataFrame
        The DataFrame to generate summary statistics
    by: list
        List of columns representing time periods to group by
    subset : list of str, default: None
        The list of column names to generate summary statistics. If None, all
        columns are used.
    percentiles : tuple of float, default: (.01, .05, .25, .50, .75, .95, .99)
        The list of percentiles in the table.

    Returns
    -------
    pandas.DataFrame
        The summary statistics. Each row represents one variable and each
        column represents one statistic.
    """
    if not isinstance(by, list):
        raise TypeError('by must be a list')
    if not isinstance(subset, list):
        raise TypeError('subset must be a list')
    # Select columns (only numeric ones)
    if subset is None:
        subset = list(data.select_dtypes(include=['number']).columns)
    else:
        if not set(subset).issubset(set(data.columns)):
            raise ValueError('subset is not in columns of data')
        subset = list(data[subset].select_dtypes(include=['number']).columns)
    # Sort


    @njit
    def _sumstat_wrapper(arrlist, percentiles):
        data = arrlist[0]
        nstats = len(percentiles) + 5
        nvars = data.shape[1]
        out = np.zeros((1, nvars * nstats))
        out.fill(np.nan)
        for i in range(nvars):
            out[0,nstats*i+0] = np.isfinite(data[:,i]).sum()
            out[0,nstats*i+1] = np.nanmean(data[:,i])
            out[0,nstats*i+2] = np.nanstd(data[:,i])
            out[0,nstats*i+3] = np.nanmin(data[:,i])
            for j in range(len(percentiles)):
                out[0,nstats*i+j+4] = np.nanquantile(data[:,i], percentiles[j])
            out[0,nstats*i+len(percentiles)+4] = np.nanmax(data[:,i])
        return out


    statslist = ['N', 'Mean', 'Std', 'Min'] + [
        'p' + str(int(i * 100)) for i in percentiles
    ] + ['Max']
    colout = [(x, y) for x in subset for y in statslist]
    dftmp = groupby_apply(
        data=data[by + subset].sort_values(by),
        by=by,
        func=_sumstat_wrapper,
        colargs=[subset],
        otherargs=(percentiles,),
        colout=colout,
    )
    dftmp.drop(columns=by, inplace=True)
    dftmp.columns = pd.MultiIndex.from_tuples(dftmp.columns)
    nrows = dftmp.shape[0]
    rowlist = dftmp.columns.get_level_values(0).unique()
    collist = dftmp.columns.get_level_values(1).unique()
    out = dftmp.mean().unstack(1).loc[rowlist, collist]
    # Calculate the number of observations
    out['N'] = (out['N'] * nrows).astype('int')
    return out


def format_table(data, float_digit=3):
    """Format data in table"""
    intcols = data.select_dtypes(include='int').columns
    floatcols = data.select_dtypes(include='float').columns
    out = []
    for col in data.columns:
        if col in intcols:
            out.append(data[col].apply(lambda x: f'{x:,d}'))
        elif col in floatcols:
            out.append(data[col].apply(lambda x: f'{x:.{float_digit}f}'))
        else:
            ValueError(f'Column {col} is not of type int or float')
    return pd.concat(out, axis=1)



# =============================================================================#
# Classes related to reporting output tables
# =============================================================================#

class FMResult:
    """The estimation result from a linear regression model

    Properties
    ----------
    eststo : list of dict

    Methods
    -------
    add
        Add regression results and statistics
    output
        Print regression table
    """

    def __init__(self):
        self._eststo = []

    def __getitem__(self, key):
        if key >= len(self._eststo):
            raise IndexError('list index out of range')
        else:
            return self._eststo[key]

    @property
    def eststo(self):
        """A list of dict containing estimates and stats"""
        return self._eststo

    def add(self, params, stderrs, labels, doff, stat):
        if not isinstance(params, np.ndarray):
            raise TypeError('params is not of type numpy.ndarray')
        if not isinstance(stderrs, np.ndarray):
            raise TypeError('stderrs is not of type numpy.ndarray')
        if not params.ndim == 1:
            raise TypeError('params is not 1-D')
        if not stderrs.ndim == 1:
            raise TypeError('stderrs is not 1-D')
        if not params.shape == stderrs.shape:
            raise ValueError('params and stderrs have different shapes.')
        if not isinstance(labels, list):
            raise TypeError('labels is not a list')
        for s in labels:
            if not isinstance(s, str):
                raise TypeError(f'element {s} in labels is not str')
        if not isinstance(stat, (dict, pd.Series)):
            raise TypeError('stat is not a dict or pandas.Series')
        self._eststo.append({
            'est': [PointEstimate(params[i],
                                  stderrs[i],
                                  labels[i],
                                  doff)
                    for i in range(params.shape[0])],
            'stat': stat if isinstance(stat, pd.Series) else pd.Series(stat)
        })

    def output(self, type: str, estfmt: tuple, statfmt: tuple, roworder: list,
               tex: bool=False, scale: float=1) -> pd.DataFrame:
        """Print the regression table.

        Parameters
        ----------
        type : str, either 'tstat' or 'pval'
        estfmt : 2-d tuple
        statfmt : tuple
        roworder : list
        tex : bool, default: False
        scale : float, default: 1

        Returns
        -------
        pandas.DataFrame
        """
        # ----------- Coefficients
        # Get Series-format table
        uptbl = pd.concat([
            pd.concat([x.output(type, tex, estfmt, scale)
                       for x in y['est']])
            for y in self._eststo
        ], axis=1).reindex(roworder, level=0).fillna('')
        # Further formatting
        uptbl.index.names = [None, None]
        uptbl = uptbl.reset_index()
        uptbl.loc[lambda x: x['level_1'] == type, 'level_0'] = ''
        uptbl = uptbl.drop(columns=['level_1']) \
            .rename(columns={'level_0': ''})
        # ----------- Statistics
        # Get the table
        lotbl = pd.concat([
            x['stat'] for x in self.eststo
        ], axis=1)
        assert len(statfmt) == len(lotbl.index)
        # Format numbers
        lotbl = pd.DataFrame(
            [
                [
                    ('{:'+statfmt[i]+'}').format(v)
                    for v in lotbl.iloc[i].to_numpy()
                ]
                for i in range(lotbl.shape[0])
            ],
            index=lotbl.index
        )
        lotbl.index.name = ''
        lotbl.reset_index(inplace=True)
        tbl = pd.concat([uptbl, lotbl], axis=0)
        return tbl

class PointEstimate:
    """Point estimate of one coefficient

    Properties
    ----------
    param : float
        Point estimate
    stderr : float
        Standard error of the point estimate
    label : str
        Label of the variable
    doff : int
        Degrees of freedom, used to calculate p-value
    tstat : float
        t-statistic
    pval : float
        p-value based on Student's t distribution with `doff` degrees of freedom
    nstars : int
        Number of stars
    starlabel : str
        A string containing `nstars` stars.
    """

    def __init__(self, param, stderr, label, doff):
        if not isinstance(param, float):
            raise TypeError('param is not float')
        if not isinstance(stderr, float):
            raise TypeError('stderr is not float')
        if not isinstance(label, str):
            raise TypeError('label is not float')
        if not isinstance(doff, (int, float, np.number)):
            raise TypeError('doff is not of type int, float, or numpy.number')
        self._param = param
        self._stderr = stderr
        self._label = label
        self._doff = doff
        self._tstat = param / stderr
        self._pval = 2 * (1 - tdist.cdf(np.abs(self._tstat), self._doff))
        if 0 <= self._pval <= 0.01:
            self._nstars = 3
        elif 0.01 < self._pval <= 0.05:
            self._nstars = 2
        elif 0.05 < self._pval <= 0.1:
            self._nstars = 1
        else:
            self._nstars = 0
        self._starlabel = '*' * self._nstars

    @property
    def param(self):
        """Point estimate"""
        return self._param

    @property
    def stderr(self):
        """Standard error of the point estimate"""
        return self._stderr

    @property
    def label(self):
        """Label of the variable"""
        return self._label

    @property
    def doff(self):
        """Degrees of freedom to calculate p-value"""
        return self._doff

    @property
    def tstat(self):
        """t-statistic"""
        return self._tstat

    @property
    def pval(self):
        return self._pval

    def output(self, stat, tex, outputfmt, scale):
        """Return a pandas.Series of formatted output"""
        if not (isinstance(outputfmt, tuple) and len(outputfmt) == 2):
            raise TypeError('outputfmt must be tuple of length 2')
        out = pd.Series(
            ['', ''],
            index=pd.MultiIndex.from_product(
                [[self._label], ['param', stat]])
        )
        # Parameter estimate
        paramstr = ('{:' + outputfmt[0] + '}').format(self._param*scale)
        if self._nstars > 0:
            if tex:
                paramstr += r'\sym{' + self._starlabel + '}'
            else:
                paramstr += self._starlabel
        if tex and self._param < 0:
            paramstr = paramstr.replace('-', '$-$')
        out[(self._label, 'param')] = paramstr
        # Statistic
        if stat == 'tstat':
            statstr = ('({:' + outputfmt[1] + '})').format(self._tstat)
        elif stat == 'pval':
            statstr = ('({:' + outputfmt[1] + '})').format(self._pval)
        else:
            statstr = '()'
        if tex:
            statstr = statstr.replace('-', '$-$')
        out[(self._label, stat)] = statstr
        return out


def regtablatex(data: pd.DataFrame) -> str:
    """Generate tex table from regression table.

    Parameters:
    ----------
    data : DataFrame
        The regression table, similar to the output of fmreg()

    Returns:
    ----------
    output : string
        The tex code of the table
    """
    # Check duplicated column headers
    if not len(data.columns) == len(data.columns.to_flat_index().unique()):
        raise ValueError('Duplicated columns headers detected!')
    # ---- Preparation
    data = data.reset_index(drop=True).copy()
    # Number of columns
    nrow = data.shape[0]
    ncol = data.shape[1]
    # column number of last line with coefficients or t-stats
    coefendflag = (data.iloc[:, 0] == '')[::-1].idxmax()
    # Add modifiers to negative signs and stars
    starpat = re.compile(r'\*{1,3}')
    negpat = re.compile(r'\-{1}')
    for i in range(nrow):
        for j in range(1, ncol):
            targetstr = data.iloc[i, j]
            # Add \sym{ } around stars
            matched = starpat.search(targetstr)
            if matched is not None:
                targetstr = re.sub(
                    starpat,
                    "\\\\sym{" +
                    targetstr[matched.start():matched.end()] + "}",
                    targetstr
                )
            # Add $ $ around -
            matched = negpat.search(targetstr)
            if matched is not None:
                targetstr = re.sub(negpat, "$-$", targetstr)
            data.iloc[i, j] = targetstr
    # Get max width for each column
    maxwidth = data.apply(lambda x: x.apply(lambda x: len(x)).max()).tolist()
    # Add spaces to each cell
    for j in range(ncol):
        data.iloc[:, j] = data.iloc[:, j].apply(
            lambda x: ("{:" + str(maxwidth[j]) + "s}").format(x)
        )
    # ---- Output
    strout = ""
    # Header
    strout += "{\n\\def\\sym#1{\\ifmmode^{#1}\\else\\(^{#1}\\)\\fi}\n"
    strout += "\\begin{tabular}{l*{" + str(ncol-1) + "}{c}}\n"
    # Header row
    strout += "\\toprule\n"
    nlevels = data.columns.nlevels
    # Loop over each level
    for i in range(nlevels):
        colnames = data.columns.get_level_values(i)
        # Loop over each value
        j = 0
        while j < ncol:
            # Get the number of repetitions
            k = 1
            while j+k < ncol and colnames[j+k] == colnames[j]:
                k += 1
            if j > 0:
                strout += " & "
            strout += " \\multicolumn{" + str(k) + "}{c}{" + colnames[j] + "} "
            j += k
        strout += " \\\\\n"
    strout += "\\midrule\n"
    # Coefficients and t-stats
    for i in range(nrow):
        for j in range(ncol):
            strout += data.iloc[i, j]
            if j < ncol - 1:
                strout += " & "
            else:
                strout += " \\\\\n"
        if data.iloc[i, 0].strip() == "" and i <= coefendflag:
            strout += "\\addlinespace\n"
        if i == coefendflag:
            strout += "\\midrule\n"
    # Bottom
    strout += "\\bottomrule\n\\end{tabular}\n}"

    return strout


def regtabtext(data: pd.DataFrame, showindex: bool=False, colborder: bool=False,
               title: Union[str,None]=None) -> str:
    """Generate tex table from regression table.

    Parameters:
    ----------
    data : DataFrame
        The regression table, similar to the output of get_fmreg_all(). All
        columns are str. Allow for MultiIndex up to 2 levels in columns.
    showindex : bool, default: True
        Whether or not to show index in the result.
    colborder : bool, default: False
        Whether or not to show column borders in the result.
    title : str, default: None
        Title of the table

    Returns:
    ----------
    output : string
        The text code of the table
    """
    # ---------- Check arguments
    # Check duplicated column headers
    if not len(data.columns) == len(data.columns.to_flat_index().unique()):
        raise ValueError('Duplicated columns headers detected!')
    # Columns only contain string and only two levels for MultiIndex
    if isinstance(data.columns, pd.MultiIndex):
        if not data.columns.nlevels == 2:
            raise ValueError('Columns of data does not have level 2')
        if not all(isinstance(s, str) for s in data.columns.levels[0]):
            raise TypeError('Some element of level 0 of data.columns is not str')
        if not all(isinstance(s, str) for s in data.columns.levels[1]):
            raise TypeError('Some element of level 1 of data.columns is not str')
    else:
        if not all(isinstance(s, str) for s in data.columns):
            raise TypeError('Some element of data.columns is not str')
    # If showindex, then it only contains string
    if showindex:
        if not data.index.dtype.kind == 'O':
            raise ValueError('Index is not of str type.')
    # ---- Preparation
    data = data.reset_index(drop=not showindex).copy()
    # Number of columns and rows
    nrow = data.shape[0]
    ncol = data.shape[1]
    # column number of last line with coefficients or t-stats
    coefendflag = (data.iloc[:, 0] == "")[::-1].idxmax()
    # Add space
    data = data.applymap(unstrip)
    # Get max width for each column
    maxwidth = getmaxwidth(data)
    # Add space for the first time
    data = aligntable(data, maxwidth)
    # Update max width
    maxwidth = getmaxwidth(data)
    # If columns is a MultiIndex, adjust the maxwidth
    if isinstance(data.columns, pd.MultiIndex):
        # Loop over each high-level index
        for s in data.columns.levels[0]:
            if s.strip() != '':
                s2 = unstrip(s)
                # If the width of the multicolumn text exceeds the sum of width
                # of columns exceeds the multicolumn text
                # The width of column borders is also added
                if maxwidth[s].sum() + maxwidth[s].shape[0] - 1 < len(s2):
                    # Proportionally increase all columns
                    maxwidth[[s]] += np.ceil((len(s2) -
                                              (maxwidth[s].sum() +
                                               maxwidth[s].shape[0] - 1)
                                              ) / maxwidth[s].shape[0]).astype('int')
        # Add space again
        data = aligntable(data, maxwidth)
    # ---------- Output
    strout = ""
    totalwidth = (maxwidth.sum() + data.shape[1] + 1)
    # Title
    if title is not None:
        strout += title
        strout += '\n'
    # Header
    strout += '=' * totalwidth + '\n'
    if isinstance(data.columns, pd.MultiIndex):
        # Top level
        strout += '|' + '|'.join(
            [aligntext(x,
                       maxwidth[x].sum() + maxwidth[x].shape[0] - 1,
                       'center')
             for x in data.columns.levels[0]]
        ) + '|' + '\n'
        # Bottom level
        strout += '|' + '|'.join(
            [aligntext(data.columns.get_level_values(1)[i],
                       maxwidth.iloc[i],
                       'center') for i in range(data.shape[1])]
        ) + '|' + '\n'
    else:
        strout += '|' + '|'.join(data.columns) + '|' + '\n'
    strout += '-' * totalwidth + '\n'
    # Coefficients and t-stats
    for i in range(nrow):
        strout += '|' + '|'.join(
            [data.iloc[i, j] for j in range(ncol)]
        ) + '|' + '\n'
        # Add line separating coefficients and other statistics
        if i == coefendflag:
            strout += '-' * totalwidth + '\n'
    strout += '=' * totalwidth + '\n'
    if not colborder:
        strout = strout.replace('|', ' ')
    return strout


def aligntable(data, maxwidth):
    ncol = data.shape[1]
    # Add spaces to each cell
    for j in range(ncol):
        if j == 0:
            # Align left for the first column
            data.iloc[:, j] = data.iloc[:, j].apply(
                lambda x: aligntext(x, maxwidth.iloc[j], 'left')
            )
        else:
            # Align centered for the others
            data.iloc[:, j] = data.iloc[:, j].apply(
                lambda x: aligntext(x, maxwidth.iloc[j], 'center')
            )
    return data


def aligntext(text, width, align):
    """Align text with specific width and direction
    """
    if not isinstance(text, str):
        raise ValueError('text is not of type str.')
    if not int(width) == width:
        raise ValueError('width is not an integer')
    if not width >= len(text):
        raise ValueError(f'width ({width}) is smaller than the width of text ({text})')
    if text == '':
        return ' ' * width
    if align == 'left':
        alignsym = '<'
    elif align == 'right':
        alignsym = '>'
    elif align == 'center':
        alignsym = '^'
    else:
        raise ValueError(f'Invalid align type: {align}')
    outstr = ('{:' + alignsym + str(width) + 's}').format(text)
    return outstr


def unstrip(text):
    if len(text) == 0:
        return ''
    if text[0] != ' ':
        text = ' ' + text
    if text[-1] != ' ':
        text = text + ' '
    return text


def getmaxwidth(data):
    return data.apply(
        lambda col: col.apply(lambda cell: len(cell)).max()
    )


# cut portfolios into groups
@njit
def qcut_jit(x, q):
    return np.floor(np.searchsorted(np.sort(x), x, side='right') * q / len(x) - 1e-12) + 1

# # Univariate portfolio sorting
# def uniportsort(
#         data: pd.DataFrame, retvar, sortvar, ngroups: int, weight, maxlag: int,
#         fdata=None, fmodel=None, scale=1
# ):
#     """Generate tex table from regression table.
#
#     Parameters:
#     ----------
#     data : pandas.DataFrame
#         Panel data containing all variables needed, e.g. returns, weights, and
#         variables to be sorted. Its index must be a MultiIndex with level 0
#         representing entities and level 1 representing time periods
#     retvar:
#         Column of data as the return
#     sortvar:
#         Column of data to be sorted on
#     ngroups: int
#         Number of groups (positive integer)
#     weight:
#         Column of data to calculate weighted average returns
#     maxlag: int
#         Maximum lag for Newey-West correction; negative for non-adjustment
#     fdata: pd.DataFrame, default: None
#         Table containing factors to adjusted return. Its index must be
#         consistent with level 1 of the data.index. The user should guarantee
#         that factors in fdata and returns in the column retvar of data are
#         consistent (both are in decimal or percentage). None if no adjustment
#         to be reported. fdata and fmodel must either both be None or both be
#         of the respective types.
#     fmodel: dict, default: None
#         A dict of keys (model names) and values (list of columns of fdata as
#         factors) to adjust return. None if no adjustment to be reported.
#     scale: float, default 1
#         Scale to be multiplied by the return and factors.
#     reverse: bool, default False
#         By default, the
#     Returns:
#     ----------
#
#     """
#     # Check arguments
#     if not isinstance(data, pd.DataFrame):
#         raise TypeError('data is not a pandas.DataFrame')
#     if retvar not in data.columns:
#         raise ValueError('retvar is not a column of data')
#     if sortvar not in data.columns:
#         raise ValueError('sortvar is not a column of data')
#     if not (isinstance(ngroups, int) and ngroups > 0):
#         raise TypeError('ngroups must be a positive integer')
#     if not (isinstance(ngroups, int) and ngroups > 0):
#         raise TypeError('ngroups must be a positive integer')
#     if weight not in data.columns:
#         raise ValueError('weight is not a column of data')
#     if not isinstance(maxlag, int):
#         raise TypeError('maxlag must be an integer')
#     if not ((fmodel is None and fdata is None) or (
#             isinstance(fdata, pd.DataFrame) and isinstance(fmodel, dict)
#     )):
#         raise TypeError('fdata and fmodel must either both be None or not be None')
#     shortlong = modelsetup['shortlong']
#     maxlag = modelsetup['maxlag']
#     fmodel = modelsetup['fmodel']
#     fdata = modelsetup['fdata']
#     pct = modelsetup['pct']
#     # Long-short or short-long
#     if shortlong:
#         lsvar = f'P1-P{ngroups}'
#     else:
#         lsvar = f'P{ngroups}-P1'
#     # Merge sort variable with returns
#     if weight is None:
#         dfr = pd.concat(
#             [rdata[retvar], sdata[sortvar].rename(sortvar)], axis=1
#         ).dropna(axis=0, subset=[sortvar])
#     else:
#         dfr = pd.concat(
#             [rdata[[retvar, weight]], sdata[sortvar].rename(sortvar)], axis=1
#         ).dropna(axis=0, subset=[sortvar])
#     # The user should ensure that the correct model is sent to the function
#     # if weight is None:
#     #     if modelname is None:
#     #         dfr = pd.concat(
#     #             [rdata[retvar], sdata[sortvar].rename(sortvar)],
#     #             axis=1
#     #         ).dropna(axis=0, subset=[sortvar])
#     #     else:
#     #         dfr = pd.concat(
#     #             [rdata[retvar], sdata[(modelname, sortvar)].rename(sortvar)],
#     #             axis=1
#     #         ).dropna(axis=0, subset=[sortvar])
#     # else:
#     #     if modelname is None:
#     #         dfr = pd.concat(
#     #             [rdata[[retvar,weight]], sdata[sortvar].rename(sortvar)],
#     #             axis=1
#     #         ).dropna(axis=0, subset=[sortvar])
#     #     else:
#     #         dfr = pd.concat(
#     #             [rdata[[retvar,weight]], sdata[(modelname, sortvar)].rename(sortvar)], axis=1
#     #         ).dropna(axis=0, subset=[sortvar])
#     # Get ranks
#     dfr['rnk'] = dfr.groupby(timevar)[sortvar].transform(lambda x: qcut_jit(x.values, q=ngroups)).astype('int')
#     # Calculate portfolio return
#     if weight is None:
#         dfp = dfr.reset_index().groupby([timevar, 'rnk']).apply(
#             lambda x: np.mean(x[retvar])
#         ).to_frame(name=retvar)
#     else:
#         dfp = dfr.reset_index().groupby([timevar, 'rnk']).apply(
#             lambda x: np.average(x[retvar], weights=x[weight])
#         ).to_frame(name=retvar)
#     # Calculate average sorted variable
#     dfs = dfr.reset_index().groupby([timevar, 'rnk']).apply(
#         lambda x: np.mean(x[sortvar])
#     ).to_frame(name=sortvar)
#     del dfr
#     # Transpose
#     dfp = dfp.reset_index().pivot(index=timevar, columns='rnk', values=retvar).rename(
#         columns=lambda x: 'P' + str(x)
#     )
#     dfs = dfs.reset_index().pivot(index=timevar, columns='rnk', values=sortvar).rename(
#         columns=lambda x: 'P' + str(x)
#     )
#     # Average sort variable
#     res_sortvar = dfs.mean().apply(lambda x: "{:5.4f}".format(x)).rename('fconst').to_frame().assign(ftstat="").stack().rename(r'$\beta_{\mathrm{'+sortvar+r'}}$')
#     # Long-short portfolio return
#     if shortlong:
#         dfp[lsvar] = dfp['P1'] - dfp[f'P{ngroups}']
#     else:
#         dfp[lsvar] = dfp[f'P{ngroups}'] - dfp['P1']
#     # Calculate average raw return and t-stat for each portfolio
#     dfavgp = pd.concat([_get_ave_ret(x[1], maxlag) for x in dfp.items()], axis=1, keys=dfp.columns)
#     dfavgp = dfavgp.T
#     # Format average return and t-stat
#     dfavgp['fconst'] = dfavgp[['const','t','pval']].apply(lambda x: _format_const(x, 'const', 'pval', pct), axis=1)
#     dfavgp['ftstat'] = dfavgp[['t']].apply(lambda x: '({:4.2f})'.format(x['t']), axis=1)
#     # Generate output column for average return
#     res_avgret = dfavgp[['fconst','ftstat']].stack().rename('Average return')
#     # Calculate alpha or not
#     if fmodel is None or fdata is None:
#         res_out = pd.concat([res_sortvar, res_avgret], axis=1).fillna(value="").reindex(res_avgret.index).reset_index().drop(columns='level_1')
#     else:
#         # Calculate and format alpha
#         res_alpha = pd.concat([_get_formated_alpha(dfp, fdata.reindex(dfp.index), fvars, maxlag, pct) for fname, fvars in fmodel.items()], axis=1, keys=fmodel.keys())
#         res_out = pd.concat([res_sortvar, res_avgret, res_alpha], axis=1).fillna(value="").reindex(res_avgret.index).reset_index().drop(columns='level_1')
#     # Output
#     res_out['rnk'] = res_out['rnk'].where(np.arange((ngroups+1)*2) % 2 == 0, other="")
#     res_out.rename(columns={'rnk':'Portfolio'}, inplace=True)
#     return res_out
#
#
# def biportsortall(modelsetup, cdata, ctrlvarlst, fvars):
#     res_out = pd.concat(
#         map(lambda x: biportsort1(modelsetup, cdata, x, fvars), ctrlvarlst), axis=1, keys=ctrlvarlst
#     ).reset_index().drop(columns='level_1')
#     res_out['rnk_sort'] = res_out['rnk_sort'].where(np.arange((modelsetup['ngroups']+1)*2) % 2 == 0, other="")
#     res_out.rename(columns={'rnk_sort':'Portfolio'}, inplace=True)
#     return res_out
#
#
# def biportsort1(modelsetup, cdata, ctrlvar, fvars):
#     rdata = modelsetup['rdata']
#     sdata = modelsetup['sdata']
#     timevar = modelsetup['timevar']
#     # modelname = modelsetup['modelname']
#     sortvar = modelsetup['sortvar']
#     retvar = modelsetup['retvar']
#     ngroups = modelsetup['ngroups']
#     weight = modelsetup['weight']
#     shortlong = modelsetup['shortlong']
#     maxlag = modelsetup['maxlag']
#     fdata = modelsetup['fdata']
#     pct = modelsetup['pct']
#     # Long-short or short-long
#     if shortlong:
#         lsvar = f'P1-P{ngroups}'
#     else:
#         lsvar = f'P{ngroups}-P1'
#     # Merge sort variable with returns
#     if weight is None:
#         dfr = pd.concat(
#             [rdata[retvar], cdata[ctrlvar], sdata[sortvar].rename(sortvar)], axis=1
#         ).dropna(axis=0, subset=[ctrlvar, sortvar])
#     else:
#         dfr = pd.concat(
#             [rdata[[retvar, weight]], cdata[ctrlvar], sdata[sortvar].rename(sortvar)], axis=1
#         ).dropna(axis=0, subset=[ctrlvar, sortvar])
#     # if weight is None:
#     #     if modelname is None:
#     #         dfr = pd.concat(
#     #             [rdata[retvar], cdata[ctrlvar], sdata[sortvar].rename(sortvar)],
#     #             axis=1
#     #         ).dropna(axis=0, subset=[ctrlvar, sortvar])
#     #     else:
#     #         dfr = pd.concat(
#     #             [rdata[retvar], cdata[ctrlvar], sdata[(modelname, sortvar)].rename(sortvar)],
#     #             axis=1
#     #         ).dropna(axis=0, subset=[ctrlvar, sortvar])
#     # else:
#     #     if modelname is None:
#     #         dfr = pd.concat(
#     #             [rdata[[retvar,weight]], cdata[ctrlvar], sdata[sortvar].rename(sortvar)],
#     #             axis=1
#     #         ).dropna(axis=0, subset=[ctrlvar, sortvar])
#     #     else:
#     #         dfr = pd.concat(
#     #             [rdata[[retvar,weight]], cdata[ctrlvar], sdata[(modelname, sortvar)].rename(sortvar)], axis=1
#     #         ).dropna(axis=0, subset=[ctrlvar, sortvar])
#     # Get ctrl variable rank
#     dfr['rnk_ctrl'] = dfr.groupby(timevar)[ctrlvar].transform(
#         lambda x: qcut_jit(x.values, q=ngroups)
#     ).astype('int')
#     # Get sort variable rank controlling for ctrl variable
#     # dfr['rnk_sort'] = dfr.groupby([timevar, 'rnk_ctrl'])[sortvar].transform(
#     #     lambda x: pd.qcut(x, q=ngroups, labels=np.arange(1, ngroups+1))
#     # ).astype('int')
#     dfr['rnk_sort'] = dfr.groupby([timevar, 'rnk_ctrl'])[sortvar].transform(
#         lambda x: qcut_jit(x.values, q=ngroups)
#     ).astype('int')
#     # Calculate portfolio return
#     if weight is None:
#         dfp = dfr.reset_index().groupby([timevar, 'rnk_ctrl', 'rnk_sort']).agg({retvar: 'mean'})
#     else:
#         dfr['__rw__'] = dfr[retvar].values * dfr[weight].values
#         dfr = dfr.reset_index().groupby([timevar, 'rnk_ctrl', 'rnk_sort']).agg({weight: 'sum', '__rw__': 'sum'})
#         dfp = pd.DataFrame(dfr['__rw__'].values / dfr[weight].values, index=dfr.index, columns=[retvar])
#     del dfr
#     #-------- Output 1: Average across ctrl ranks
#     # Calculate average portfolio return across all ctrl ranks
#     dfp = dfp.groupby(['date_mn', 'rnk_sort'])[[retvar]].agg('mean')
#     # Transpose
#     dfp = dfp.reset_index().pivot(index=timevar, columns='rnk_sort', values=retvar).rename(
#         columns=lambda x: 'P' + str(x)
#     )
#     # Long-short portfolio return
#     if shortlong:
#         dfp[lsvar] = dfp['P1'] - dfp[f'P{ngroups}']
#     else:
#         dfp[lsvar] = dfp[f'P{ngroups}'] - dfp['P1']
#     # Calculate average return or alpha
#     if fvars is None:
#         # Calculate average raw return and t-stat for each portfolio
#         dfavgp = pd.concat([_get_ave_ret(x[1], maxlag) for x in dfp.items()], axis=1, keys=dfp.columns)
#         dfavgp = dfavgp.T
#         # Format average return and t-stat
#         dfavgp['fconst'] = dfavgp[['const','t','pval']].apply(lambda x: _format_const(x, 'const', 'pval', pct), axis=1)
#         dfavgp['ftstat'] = dfavgp[['t']].apply(lambda x: '({:4.2f})'.format(x['t']), axis=1)
#         # Generate output column for average return
#         res_out = dfavgp[['fconst','ftstat']].stack().rename(ctrlvar)
#     else:
#         res_out = _get_formated_alpha(dfp, fdata.reindex(dfp.index), fvars, maxlag).rename(ctrlvar)
#     return res_out


def groupby_wavg(data: pd.DataFrame, bys: list, var, weight) -> pd.Series:
    """Calculate weighted average of varlist by groups

    Parameters
    ----------
    data: pd.DataFrame
        The input table
    bys: list
        Columns of data to be grouped by
    var:
        Column of data to calculate weighted average
    weight:
        Column of data as weight

    Returns
    -------
    pd.Series
        With index same as bys, the value is the by-group average of var
        weighted by weight
    """
    if isinstance(var, str):
        xyvar = var + '_xy'
    else:
        xyvar = 'xy'
    # Remove missing values
    datatmp = data.dropna(subset=[var, weight])
    outvw = datatmp.assign(
        **{xyvar: lambda x: x[var]*x[weight]}
    ).groupby(bys)[[xyvar, weight]].sum()
    outvw[var] = outvw[xyvar] / outvw[weight]
    return outvw[var]


# Estimate factor alpha given factor data (fdata), factor variables (fvars)
def _get_alpha(x, fdata, fvars, maxlag):
    exog = fdata[fvars].dropna(how='any').assign(const=1)
    endog = x.reindex(exog.index)
    mod = OLS(endog, exog).fit()
    res = pd.Series([mod.params['const'],
                     np.sqrt(cov_hac(mod, nlags=maxlag)[-1,-1]),
                     mod.df_resid], index=['const','se','df'])
    res['t'] = res['const'] / res['se']
    res['pval'] = 2 * (1 - scipy.stats.t.cdf(abs(res['t']), res['df']))
    return res


# Calculate alpha and format columns for each factor model (prdata: portfolio return data)
def _get_formated_alpha(prdata, fdata, fvars, maxlag, pct):
    dfalpha = pd.concat([_get_alpha(x[1], fdata, fvars, maxlag) for x in prdata.items()], axis=1, keys=prdata.columns)
    dfalpha = dfalpha.T
    # Format alpha
    dfalpha['fconst'] = dfalpha[['const','t','pval']].apply(lambda x: _format_const(x, 'const', 'pval', pct), axis=1)
    dfalpha['ftstat'] = dfalpha[['t']].apply(lambda x: '({:4.2f})'.format(x['t']), axis=1)
    return dfalpha[['fconst','ftstat']].stack()


# Format point estimates with stars based on pval
def _format_const(estseries, paramvar, pvalvar, scale, format='9.2f'):
    """Format point estimates with stars based on p-value.

    Parameters
    ----------
    estseries : pandas.Series
        A series of parameter estimates and p-values, such that
        estseries[parameter] is the point estimate and estseries[pvalvar] is
        the corresponding p-value.
    paramvar : str
    pvalvar : str
    scale : int
        Multiplier to apply to the point estimate.
    format : str, default: '9.2f'
        Format of the point estimate

    Returns
    -------
    str
    """
    if estseries[pvalvar] > 0.1:
        return ('{:'+format+'}').format(estseries[paramvar] * scale)
    elif (0.05 < estseries[pvalvar]) and (estseries[pvalvar] <= 0.1):
        return ('{:'+format+'}*').format(estseries[paramvar] * scale)
    elif (0.01 < estseries[pvalvar]) and (estseries[pvalvar] <= 0.05):
        return ('{:'+format+'}**').format(estseries[paramvar] * scale)
    elif (0 <= estseries[pvalvar]) and (estseries[pvalvar] <= 0.01):
        return ('{:'+format+'}***').format(estseries[paramvar] * scale)


# Estimate average return and t-stat for each column (portfolio) with Newey-West standard error
def _get_ave_ret(x, maxlag):
    exog = np.ones(x.shape)
    mod = OLS(x, exog).fit()
    res = pd.Series(
        [mod.params.values[0], np.sqrt(cov_hac(mod, nlags=maxlag))[0][0], mod.df_resid],
        index=['const', 'se', 'df']
    )
    res['t'] = res['const'] / res['se']
    res['pval'] = 2 * (1 - tdist.cdf(abs(res['t']), res['df']))
    return res


def get_port_ret(
        data: pd.DataFrame,
        nq: Union[int, List[int]],
        timevar,
        retvar,
        rnkvars: list,
        rnkvarnames: list,
        wvar=None,
        dep=False,
        retfull=False,
):
    """Construct equal-weighted and/or value-weighted portfolio returns sorted
    on different variables, dependently or independently, from a DataFrame
    containing panel returns of securities for different times.

    Parameters
    ----------
    data : DataFrame
        A panel (long-type or stacked) data of returns for different securities
        at different times.
    nq : int or list of int
        Number(s) of portfolios for each rank variable. The order should be
        consistent with `rnkvars`. If it is an integer, the number is applied
        for all rank variables.
    timevar : str
        Name of the variable representing time. This variable determines the
        time variable in the output.
    retvar : str
        Name of the variable representing security returns.
    rnkvars : list of str
        Name(s) of variables to rank on.
    rnkvarnames : list of str
        Name(s) of the created variable representing different portfolios
    wvar : str or None, default None
        Name of the variable representing portfolio weights. If it is None,
        only equal-weighted returns are calcualted
    dep : bool, default False
        Whether or not variable sorts are dependent. If `dep=True`, the order
        of variable names in `rnkvars` indicates the order of dependent sorting.
    retfull : bool, default False
        If True, a tuple will be returned from the function. The first element
        contains portfolios while the second one contains the full input data
        set with rank variables.

    Returns:
    ----------
    output : DataFrame
        A DataFrame containing a time-series of returns for different portfolios
    data : DataFrame
        The input data adding portfolio numbers
    """
    data = data.copy()
    # If nq is an integer, equivalent to nq <- [nq, ..., nq] with len(rnkvars)
    if np.ndim(nq) == 0:
        nq = [nq] * len(rnkvars)
    # Remove NaN
    data = data.dropna(subset=rnkvars+[retvar], how='any')
    # Generate group-by variable list
    by_list = [[timevar]]
    if dep:
        # Successively add each rank variable
        for v in rnkvarnames:
            by_list.append(by_list[-1] + [v])
        del by_list[len(rnkvars)]
    else:
        by_list = by_list * len(rnkvars)
    # Obtain portfolio ranks
    for i in range(len(by_list)):
        rnkvar = rnkvars[i]
        rnkvarname = rnkvarnames[i]
        data = data.groupby(by_list[i]).apply(
            lambda x: _get_qcut(x, nq[i], rnkvar, rnkvarname)
        ).reset_index(drop=True)
    # Equal-weighted return
    out = data.groupby(
        [timevar] + rnkvarnames
    )[retvar].mean().to_frame('ew_ret')
    # Value-weighted return
    if wvar is not None:
        outvw = groupby_wavg(
            data, [timevar] + rnkvarnames, retvar, wvar
        ).to_frame('vw_ret')
        out = out.join(outvw)
    # Nobs
    out = out.join(
        data.groupby([timevar] + rnkvarnames)[retvar].count().to_frame('cnt')
    ).reset_index(drop=False)
    if retfull:
        return out, data
    else:
        return out


def get_qcut(x, q, rnkvar, rnkvarname):
    """
    Similar to pd.qcut, but allow for closed interval and duplicate observations
    with multiple closed interval matched. This function is to be used in a
    `pandas.groupby()` function.

    Parameters
    ----------
    x : DataFrame
        The input array, typically a dataframe or chunk of a dataframe split by a `groupby()` function
    q : int
        Number of portfolios (e.g. 5 for quintile)
    rnkvar : str
        Name of the variable to sort on
    rnkvarname : str
        Name of the created variable representing different portfolios

    Returns
    -------
    out : DataFrame
        The output DataFrame is the same as the input `x` except that:

        * A new variable named `rnkvarname` is added representing intervals (if `retinterval=True`)
        or portfolio numbers (if `retinterval=False`).
        * Observations (rows) corresponding to multiple intervals are duplicated with the multiple intervals
        (or numbers) assigned to `rnkvarname`.

    """
    return _get_qcut(x, q, rnkvar, rnkvarname)

def _get_qcut(x, q, rnkvar, rnkvarname):
    """
    Similar to pd.qcut, but allow for closed interval and duplicate observations
    with multiple closed interval matched. This function is to be used in a
    `pandas.groupby()` function.

    Parameters
    ----------
    x : DataFrame
        The input array, typically a dataframe or chunk of a dataframe split by a `groupby()` function
    q : int
        Number of portfolios (e.g. 5 for quintile)
    rnkvar : str
        Name of the variable to sort on
    rnkvarname : str
        Name of the created variable representing different portfolios

    Returns
    -------
    out : DataFrame
        The output DataFrame is the same as the input `x` except that:

        * A new variable named `rnkvarname` is added representing intervals (if `retinterval=True`)
        or portfolio numbers (if `retinterval=False`).
        * Observations (rows) corresponding to multiple intervals are duplicated with the multiple intervals
        (or numbers) assigned to `rnkvarname`.

    """
    # Quantiles
    quantiles = np.linspace(0, 1, q + 1)
    bins = x[rnkvar].quantile(quantiles).values
    # Adjust the first and last endpoint
    bins[0] = bins[0] - (bins[1] - bins[0]) * 1e-2
    bins[-1] = bins[-1] + (bins[-1] - bins[-2]) * 1e-2
    # Intervals (seems to be unnecessary to use pd.Interval, which is slow)
    intervals = np.vstack((bins[:-1], bins[1:])).T
    interval = intervals[0]

    x_dup = pd.concat([_get_matched_data(intervals[i, :], i, x, rnkvar) for i in range(intervals.shape[0])], axis=0)
    # Drop variables
    x_dup = x_dup.drop(columns=['__interval_l__', '__interval_r__']).rename(columns={'__portnumber__': rnkvarname})
    return x_dup


# For each interval, find all obs (boolean array) whose rnkvar is in the interval
def _get_matched_data(interval, intervalno, x, rnkvar):
    # Get all obs where rnkvar is within an interval
    x_tmp = x.loc[lambda y: (interval[0] <= y[rnkvar].values) & (y[rnkvar].values <= interval[1])].copy()
    # Assign interval and number
    x_tmp['__interval_l__'] = interval[0]
    x_tmp['__interval_r__'] = interval[1]
    x_tmp['__portnumber__'] = intervalno + 1

    return x_tmp



from tabulate import tabulate
import pandas as pd
import os
from datetime import datetime
from typing import Union

def dictout(dict: dict) -> str:
    """Print a dictionary in text format
    """
    dicttmp = dict.copy()
    for k in dicttmp:
        if not isinstance(dicttmp[k], list):
            dicttmp[k] = [dicttmp[k]]
    return tabulate(dicttmp, tablefmt='psql', headers='keys')

def tabout(data: pd.DataFrame, type: str='reg', showindex: bool=False,
           floatfmt: Union[str,None]=None, suffix: str='\n') -> str:
    """Print a table in text format
    """
    if showindex:
        if data.index.name is None:
            data.index.name = ''
        data = data.reset_index(drop=False)
    # Supplement other floatfmt
    if isinstance(floatfmt, tuple) \
            and len(floatfmt) < data.shape[1]+showindex:
        floatfmt += (floatfmt[-1],) * (data.shape[1]+showindex-len(floatfmt))
    elif floatfmt is None:
        # Default is to use '.0f' for integer and '.3f' for float
        floatfmt = ()
        for v in data:
            if data[v].dtype.kind in 'biu':
                floatfmt += ('.0f',)
            elif data[v].dtype.kind in 'f':
                floatfmt += ('.3f',)
            else:
                floatfmt += ('',)
    if type == 'reg':
        data.columns = data.columns.str.replace(r'level_\d*', '', regex=True)
        return tabulate(
            data, 
            headers=data.columns,
            showindex=False,
            tablefmt='psql',
            floatfmt=floatfmt,
            colalign=('left',) + ('center',) * (data.shape[1]-1),
            ) + suffix
    else:
        raise ValueError('Invalid type: ' + type)

class logger:
    def __init__(self, file: str, overwrite: str='replace'):
        """Create a logger

        Parameters
        ----------
        file : str
            Name (and path) of the log file
        overwrite : str
            Possible values are `'replace'' and `'append''
        """
        # Open file and get the handler
        if overwrite == 'replace':
            self.filehandle = open(file, 'w')
        elif overwrite == 'append':
            self.filehandle = open(file, 'a')
        # Print file header
        self.logheader()
        
    def logheader(self):
        self.log('='*100)
        self.log(' File path: ' + self.filehandle.name)
        self.log('Created at: ' + datetime.now().strftime(r'%Y-%m-%d %H:%M:%S'))
        self.log('='*100)
        return None

    def log(self, text: str, suffix: str='\n'):
        """Print text to the log file, a newline is automatically added

        Parameters
        ----------
        text : str
        suffix : str
        """
        self.filehandle.write(text + suffix)
        self.filehandle.flush()
        os.fsync(self.filehandle)

    def close(self):
        """Close log file
        """
        self.filehandle.close()
    
    def __del__(self):
        self.filehandle.close()


import pandas as pd
from typing import Union

def rename(data: pd.DataFrame, inplace: bool=False, **kwargs) -> Union[pd.DataFrame,None]:
    """Rename columns and return the new DataFrame

    Parameters
    ----------
    data : pandas.DataFrame
    **kwargs
        Format: `oldname=newname`.

    Returns
    -------
    pandas.DataFrame
    """

    if inplace:
        data.rename(columns=kwargs, inplace=True)
        return None
    else:
        return data.rename(columns=kwargs)

def dropvar(data: pd.DataFrame, subset: Union[list,str], inplace: bool=False
            ) -> Union[pd.DataFrame,None]:
    """Drop a list of variables and return the new dataframe

    Parameters
    ----------
    data : pandas.DataFrame
    subset : list or str
    inplace : bool, default: False

    Returns
    -------
    pandas.DataFrame
    """
    if isinstance(subset, str):
        subset = subset.replace('  ', ' ').split(' ')
        assert set(subset).issubset(set(data.columns)), 'Invalid subset'
    if inplace:
        data.drop(columns=subset, inplace=True)
        return None
    else:
        return data.drop(columns=subset)


def keepna(data: pd.DataFrame, subset: Union[list,None]=None, how='any') -> pd.DataFrame:
    """Keep rows where specific variables are missing

    Parameters
    ----------
    data : pandas.DataFrame
    subset : list
    how : str, default: 'all'
        The value can be 'all' or 'any'

    Returns
    -------
    pandas.DataFrame
    """
    if subset is None:
        subset = data.columns
    else:
        assert set(subset).issubset(set(data.columns))
    if how == 'any':
        return data.loc[lambda x: x[subset].isna().any(axis=1)]
    elif how == 'all':
        return data.loc[lambda x: x[subset].isna().all(axis=1)]
    else:
        raise ValueError('Invalid parameter how: ' + how)



from setuptools import setup

setup(name='pyempfin',
      version='0.1.1',
      description='Helper functions for empirical finance research',
      author='Francis Cong',
      license='MIT',
      packages=['pyempfin'],
      install_requires=[
          'numpy',
          'pandas',
          'numba',
          'tabulate',
          'joblib',
          'scipy'
      ],
      zip_safe=False)
